#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import system
from . import legacy

METHOD_NAME_VALUE = "method_name"
""" The method name value """

def load_plugin(lazy_loading = False, metadata_enabled = True):
    """
    Decorator for the initial load of the plugin.

    :type lazy_loading: bool
    :param lazy_loading: The boolean flag to set the type of loading for the plugin.
    :type metadata_enabled: bool
    :param metadata_enabled: The boolean flag to set the metadata availability.
    :rtype: Function
    :return: The created decorator.
    """

    def create_decorator_interceptor(function):
        """
        Creates a decorator interceptor, that intercepts the normal function call.

        :type function: Function
        :param function: The callback function.
        """

        def decorator_interceptor(*args, **kwargs):
            """
            The interceptor function for the load allowed decorator.
            """

            # calls the callback function
            function(*args, **kwargs)

            # retrieves the plugin instance
            plugin = args[0]

            # retrieves the metadata map for the current plugin
            metadata_map = function.metadata

            # sets the metadata map in the plugin instance
            plugin.metadata_map = metadata_map

        # returns the decorator interceptor
        return decorator_interceptor

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the load allowed decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The decorator interceptor function.
        """

        # starts the matadata map for
        # the current function
        function.metadata = {}

        # sets the current load plugin function
        # for the load allowed reference
        load_plugin.current = function

        # creates the decorator interceptor with the given function
        decorator_interceptor_function = create_decorator_interceptor(function)

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def plugin_meta_information(metadata_key, metadata_values = {}):
    """
    Decorator that sets a defined metadata value in a plugin.

    :type metadata_key: String
    :param metadata_key: The metadata key to process the metadata.
    :type metadata_values: Dictionary
    :param metadata_values: The dictionary containing the metadata values.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the plugin meta information decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The decorator interceptor function.
        """

        # retrieves the current load plugin function
        load_plugin_current = load_plugin.current

        # retrieves the metadata plugin map
        metadata_plugin_map = load_plugin_current.metadata

        # in case the metadata key does not exist in the metadata plugin map
        if not metadata_key in metadata_plugin_map:
            # creates a new list for the metadata key
            metadata_plugin_map[metadata_key] = []

        # retrieves the function name
        function_name = function.__name__

        # sets the value for the metadata method name
        metadata_values[METHOD_NAME_VALUE] = function_name

        # appends the metadata values to the list of values with the same metadata key
        metadata_plugin_map[metadata_key].append(metadata_values)

        # returns the function
        return function

    # returns the created decorator
    return decorator

def load_allowed(function):
    """
    Decorator that loads the allowed plugins into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the load allowed decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the allowed functions map
        allowed_functions_map = function.allowed_functions_map

        # unpacks the various arguments for the function
        original_plugin = args[0]
        allowed_plugin = args[1]
        capability = args[2]

        # iterates over all the capabilities in the allowed
        # functions map
        for capability_key in allowed_functions_map:
            # checks if the capability is a valid capability for the plugin
            # currently being handled and skips the current iteration
            # in case it's not
            if not system.is_capability_or_sub_capability(capability_key, capability):
                # continues the loop
                continue

            # retrieves the capability function name from the allowed
            # functions map
            capability_function = allowed_functions_map[capability]
            capability_function_name = capability_function.__name__

            # retrieves the capability method from the original plugin and
            # handles the capability
            capability_method = getattr(original_plugin, capability_function_name)
            capability_method(allowed_plugin, capability)

    # starts the allowed functions map for
    # the current function
    function.allowed_functions_map = {}

    # sets the current load allowed function
    # for the load allowed reference
    load_allowed.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def load_allowed_capability(capability, load_plugin = False):
    """
    Decorator that marks a method for loading of allowed plugins.

    :type capability: String
    :param capability: The name of the capability to be loaded.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the load allowed capability decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # retrieves the current load allowed function
        load_allowed_current = load_allowed.current

        # sets the current function for capability handling
        # in the current function
        load_allowed_current.allowed_functions_map[capability] = function

        # in case the load plugin test should be made before loading
        # the capability
        if load_plugin:
            # creates the new decorator interceptor function
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is set
        else:
            # sets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def unload_allowed(function):
    """
    Decorator that unloads the allowed plugins into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the unload allowed decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the allowed functions map
        allowed_functions_map = function.allowed_functions_map

        # unpacks the various arguments for the function
        original_plugin = args[0]
        allowed_plugin = args[1]
        capability = args[2]

        # iterates over all the capabilities in the allowed
        # functions map
        for capability_key in allowed_functions_map:
            # checks if the capability is a valid capability for the plugin
            # currently being handled and skips the current iteration
            # in case it's not
            if not system.is_capability_or_sub_capability(capability_key, capability):
                # continues the loop
                continue

            # retrieves the capability function name from the allowed
            # functions map
            capability_function = allowed_functions_map[capability]
            capability_function_name = capability_function.__name__

            # retrieves the capability method from the original plugin and
            # handles the capability
            capability_method = getattr(original_plugin, capability_function_name)
            capability_method(allowed_plugin, capability)

    # starts the allowed functions map for
    # the current function
    function.allowed_functions_map = {}

    # sets the current unload allowed function
    # for the unload allowed reference
    unload_allowed.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def unload_allowed_capability(capability, load_plugin = False):
    """
    Decorator that marks a method for unloading of allowed plugins.

    :type capability: String
    :param capability: The name of the capability to be unloaded.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the unload allowed capability decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # retrieves the current unload allowed function
        unload_allowed_current = unload_allowed.current

        # sets the current function for capability handling
        # in the current function
        unload_allowed_current.allowed_functions_map[capability] = function

        # in case the load plugin test should be made before unloading
        # the capability
        if load_plugin:
            # creates the new decorator interceptor function
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is set
        else:
            # sets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def inject_dependencies(function):
    """
    Decorator that injects the dependencies into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the inject dependencies decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the dependency functions map
        dependency_functions_map = function.dependency_functions_map

        # retrieves the original plugin
        original_plugin = args[0]

        # retrieves the dependency plugin
        dependency_plugin = args[1]

        # creates the dependency plugin tuple
        dependency_plugin_tuple = (
            dependency_plugin.original_id,
            dependency_plugin.version
        )

        # creates the dependency plugin simple tuple
        # this tuple only contains the
        dependency_plugin_simple_tuple = (
            dependency_plugin.original_id,
            None
        )

        # tries to retrieve the set function using both the complete plugin id and
        # version approach and the version only approach
        set_function = dependency_functions_map.get(dependency_plugin_tuple, None)
        set_function = set_function or dependency_functions_map.get(dependency_plugin_simple_tuple, None)

        # in case the set function is not defined
        if not set_function:
            # returns immediately
            return

        # retrieves the set function name from the dependency functions map
        set_function = dependency_functions_map[dependency_plugin_simple_tuple]
        set_function_name = set_function.__name__

        # calls the set method from the original plugin instance
        set_method = getattr(original_plugin, set_function_name)
        set_method(dependency_plugin)

    # starts the dependency functions map for
    # the current function
    function.dependency_functions_map = {}

    # sets the current inject dependencies function
    # for the inject dependencies reference
    inject_dependencies.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def plugin_inject(plugin_id, plugin_version = None, load_plugin = False):
    """
    Decorator that marks a method for injection of a dependency.

    :type plugin_id: String
    :param plugin_id: The id of the plugin dependency to be injected.
    :type plugin_version: String
    :param plugin_version: The version of the plugin dependency to be injected.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the plugin inject decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # creates the dependency plugin tuple
        dependency_plugin_tuple = (
            plugin_id,
            plugin_version
        )

        # retrieves the current inject dependencies function
        inject_dependencies_current = inject_dependencies.current

        # sets the current function for dependency injection
        # in the current function
        inject_dependencies_current.dependency_functions_map[dependency_plugin_tuple] = function

        # in case the load plugin test should be made before injecting
        # the dependency
        if load_plugin:
            # creates the new decorator interceptor functions
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is set
        else:
            # sets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def event_handler(function):
    """
    Decorator that redirects the event handling into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the event handler decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the event handler functions map
        event_handler_functions_map = function.event_handler_functions_map

        # unpacks the function arguments
        original_plugin = args[0]
        event_name = args[1]
        all_method_args = args[1:]

        # iterates over all the events in the evens in the event handler
        # functions map
        for event_name_key in event_handler_functions_map:
            # checks if the event is a valid event for the plugin
            # currently being handled and skips the current iteration
            # in case it's not
            if not system.is_event_or_sub_event(event_name_key, event_name):
                # continues the loop
                continue

            # retrieves the event handler function from event handler functions map
            event_handler_function = event_handler_functions_map[event_name_key]

            # retrieves the function name and the event handler method from the original
            # plugin
            event_handler_function_name = event_handler_function.__name__
            event_handler_method = getattr(original_plugin, event_handler_function_name)

            # retrieves the number of arguments for the function by
            # inspecting the specification of the function, this may
            # be an expensive operation and should be used with care
            spec = legacy.getargspec(event_handler_function)
            number_arguments = len(spec.args)

            # in case the length of the arguments is insufficient
            if len(all_method_args) < number_arguments:
                # adds padding to the call arguments
                call_args = list(all_method_args) + [None]
            else:
                # maintains the calling arguments
                call_args = all_method_args

            # calls the event handler method
            event_handler_method(*call_args)

    # starts the event handler functions map for
    # the current function
    function.event_handler_functions_map = {}

    # sets the current event handler function
    # for the event handler reference
    event_handler.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def event_handler_method(event_name, load_plugin = False):
    """
    Decorator that marks a method for event handling.

    :type event_name: String
    :param event_name: The name of the event to be handled by the marked method.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the event handler method decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # retrieves the current event handler function
        event_handler_current = event_handler.current

        # sets the current function for event handling
        # in the current function
        event_handler_current.event_handler_functions_map[event_name] = function

        # in case the load plugin test should be made before handling
        # the event
        if load_plugin:
            # creates the new decorator interceptor functions
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is set
        else:
            # sets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def set_configuration_property(function):
    """
    Decorator that redirects the setting of configuration properties into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the set configuration
        property decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the set configuration property functions map
        set_configuration_property_functions_map = function.set_configuration_property_functions_map

        # unpacks the function arguments
        original_plugin = args[0]
        property_name = args[1]
        property = args[2]

        # in case the property name is not defined in the set configuration
        # property functions map
        if not property_name in set_configuration_property_functions_map:
            # returns immediately
            return

        # retrieves the set configuration property function from set configuration property functions map
        set_configuration_property_function = set_configuration_property_functions_map[property_name]

        # retrieves the function name and the set configuration property method from the original
        # plugin
        set_configuration_property_function_name = set_configuration_property_function.__name__
        set_configuration_property_method = getattr(original_plugin, set_configuration_property_function_name)

        # calls the set configuration property method
        set_configuration_property_method(property_name, property)

    # starts the set configuration property functions map for
    # the current function
    function.set_configuration_property_functions_map = {}

    # sets the current set configuration property function
    # for the set configuration property reference
    set_configuration_property.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def set_configuration_property_method(property_name, load_plugin = False):
    """
    Decorator that marks a method for set configuration property.

    :type property_name: String
    :param property_name: The name of the property to be set by the marked method.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the set configuration property decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # retrieves the current set configuration property function
        set_configuration_property_current = set_configuration_property.current

        # sets the current function for set configuration property
        # in the current function
        set_configuration_property_current.set_configuration_property_functions_map[property_name] = function

        # in case the load plugin test should be made before setting
        # the configuration property
        if load_plugin:
            # creates the new decorator interceptor functions
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is set
        else:
            # sets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def unset_configuration_property(function):
    """
    Decorator that redirects the unsetting of configuration
    properties into the defined methods.

    :type function: Function
    :param function: The function to be decorated.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the unset configuration
        property decorator.
        """

        # calls the callback function
        function(*args, **kwargs)

        # retrieves the unset configuration property functions map
        unset_configuration_property_functions_map = function.unset_configuration_property_functions_map

        # unpacks the function arguments
        original_plugin = args[0]
        property_name = args[1]

        # in case the property name is not defined in the unset configuration
        # property functions map
        if not property_name in unset_configuration_property_functions_map:
            # returns immediately
            return

        # retrieves the unset configuration property function from unset configuration property functions map
        unset_configuration_property_function = unset_configuration_property_functions_map[property_name]

        # retrieves the function name and the unset configuration property method from the original
        # plugin
        unset_configuration_property_function_name = unset_configuration_property_function.__name__
        unset_configuration_property_method = getattr(original_plugin, unset_configuration_property_function_name)

        # calls the unset configuration property method
        unset_configuration_property_method(property_name)

    # starts the unset configuration property functions map for
    # the current function
    function.unset_configuration_property_functions_map = {}

    # unsets the current unset configuration property function
    # for the unset configuration property reference
    unset_configuration_property.current = function

    # returns the decorator interceptor
    return decorator_interceptor

def unset_configuration_property_method(property_name, load_plugin = False):
    """
    Decorator that marks a method for unset configuration property.

    :type property_name: String
    :param property_name: The name of the property to be unset by the marked method.
    :type load_plugin: bool
    :param load_plugin: The load plugin flag to unset the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the unset configuration property decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The function to be decorated.
        """

        # retrieves the current unset configuration property function
        unset_configuration_property_current = unset_configuration_property.current

        # unsets the current function for unset configuration property
        # in the current function
        unset_configuration_property_current.unset_configuration_property_functions_map[property_name] = function

        # in case the load plugin test should be made before unsetting
        # the configuration property
        if load_plugin:
            # creates the new decorator interceptor functions
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        # otherwise the normal approach is unset
        else:
            # unsets the decorator interceptor function as the function
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def plugin_call(load_plugin = True):
    """
    Decorator that intercepts a plugin front-end call method.

    :type load_plugin: bool
    :param load_plugin: The load plugin flag to set the test for plugin loading.
    :rtype: Function
    :return: The created decorator.
    """

    def decorator(function, *args, **kwargs):
        """
        The decorator function for the plugin call decorator.

        :type function: Function
        :param function: The function to be decorated.
        :rtype: Function
        :return: The decorator interceptor function.
        """

        if load_plugin:
            decorator_interceptor_function = create_load_plugin_interceptor(function)
        else:
            decorator_interceptor_function = function

        # returns the interceptor to be used
        return decorator_interceptor_function

    # returns the created decorator
    return decorator

def create_load_plugin_interceptor(function):
    """
    Creates a load plugin interceptor, that loads a lazy plugin on the fly.

    :type function: Function
    :param function: The callback function.
    """

    def decorator_interceptor(*args, **kwargs):
        """
        The interceptor function for the decorator.
        """

        # retrieves the original plugin instance
        original_plugin = args[0]

        # in case the plugin is not yet loaded
        if not original_plugin.is_loaded():
            plugin_manager = original_plugin.manager
            original_plugin_id = original_plugin.id

            # in case the plugin load was unsuccessful
            if not plugin_manager.load_plugin(original_plugin_id, system.FULL_LOAD_TYPE):
                return None

        # calls the callback function
        return function(*args, **kwargs)

    # returns the decorator interceptor
    return decorator_interceptor



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import inspect

def get_instance_module_directory(instance):
    """
    Retrieves the directory path from the given
    instance value.

    :type instance: Object
    :param instance: The instance value to be used to
    retrieve the module.
    :rtype: String
    :return: The path to the directory that contains
    the module with the given instance.
    """

    # retrieves the module name
    module_name = instance.__module__

    # retrieves the module from the modules map
    module = sys.modules[module_name]

    # retrieves the module file path
    module_file_path = module.__file__

    # retrieve the module directory path
    module_directory_path = os.path.dirname(module_file_path)

    # returns the module directory path
    return module_directory_path

def get_call_module_directory(depth_level = 1):
    """
    Retrieves the directory path for the calling
    module in the given depth level of the call stack.
    This function is extremely dangerous as it may
    not work in a large range of python implementations.

    :type depth_level: int
    :param depth_level: The depth level of the call
    stack to be reached to retrieve the module.
    :rtype: String
    :return: The path to the directory that contains
    the module reference in the given level of
    the call stack.
    """

    # retrieves the current call stack
    call_stack = inspect.stack()

    # retrieves the call stack element for the
    # required depth
    call_stack_element = call_stack[depth_level + 1]

    # retrieves the call stack element's frame
    call_stack_element_frame = call_stack_element[0]

    # retrieves the (call) module from the call stack element frame
    module = inspect.getmodule(call_stack_element_frame)

    # retrieves the module file path
    module_file_path = module.__file__

    # retrieve the module directory path
    module_directory_path = os.path.dirname(module_file_path)

    # returns the module directory path
    return module_directory_path



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import time
import glob
import atexit
import signal
import logging
import warnings
import traceback
import threading

CONFIG_FILE_ENV = "COLONY_CONFIG_FILE"
""" The name of the environment variable to be used
to retrieve the path to the configuration file """

DEFAULT_CONFIG_PATH = "config/python/devel.py"
""" The path to the default configuration file to be
used in case no path is specified using the environment
variable bases strategy """

EXCLUDED_NAMES = (
    "server",
    "host",
    "port",
    "ssl",
    "key_file",
    "cer_file"
)
""" The sequence that contains the names that are considered
excluded from the auto parsing of parameters """

# retrieves the base path for the current file and uses
# it to insert it in the current system path in case it's
# not already present (required for module importing)
base_path = os.path.dirname(__file__)
if not base_path in sys.path: sys.path.insert(0, base_path)

import colony

# runs the resolution process in order to be able to retrieve
# the "real" a "best" match for the manager path, this method
# should decide between the personal and the master versions
# then ensures that the proper directory tree is created
manager_path = colony.resolve_manager(base_path)
colony.ensure_tree(manager_path)

# registers the ignore flag in the deprecation warnings so that
# no message with this kind of warning is printed (clean console)
warnings.filterwarnings("ignore", category = DeprecationWarning)

# retrieves the layout mode that is going to be used for the
# resolution of resources in the colony infra-structure
layout_mode = colony.conf("LAYOUT_MODE", "default")

# tries to retrieve the run mode from the currently set
# environment variables, in case of failure defaults to
# the default value (as expected by the specification)
run_mode = colony.conf("RUN_MODE", "development")

# tries to retrieve the prefix to be used to shorten the path
# resolution in the request handling
prefix = colony.conf("PREFIX", None)

# gathers the path to the alias file that contains JSON information
# about the mapping prefixes for the HTTP server, the file should
# contain a set of prefix to resolution prefix values
alias_path = colony.conf("ALIAS_PATH", None)
alias_path = alias_path and os.path.expanduser(alias_path)
alias_path = alias_path and os.path.normpath(alias_path)

# gathers the path to the rewrite file that contains JSON information
# about the rewrite meta information to be used at runtime to shorten
# the provided URL path values (useful under proxy redirection)
rewrite_path = colony.conf("REWRITE_PATH", None)
rewrite_path = rewrite_path and os.path.expanduser(rewrite_path)
rewrite_path = rewrite_path and os.path.normpath(rewrite_path)

# retrieves the (verbosity) level for the debugger using the provided
# configuration support, defaulting to the default level in case the
# value is not provided through configuration
level = colony.conf("LEVEL", logging.INFO)
level = colony.getLevelName(level)

# retrieves the complete set of configuration variables associated
# with the various paths to be used by colony, these are going to
# be used as the initial values for the various path lists
meta_paths = colony.conf("META_PATH", [], cast = list)
plugin_paths = colony.conf("PLUGIN_PATH", [], cast = list)

# tries to retrieve the configuration file from the environment
# variable associated in case it fails uses the default configuration
# file path then joins the "relative" file path to the base path
# and resolves it as an absolute path
config_file_path = colony.conf(CONFIG_FILE_ENV, None) or DEFAULT_CONFIG_PATH
config_file_path = os.path.join(manager_path, config_file_path)
config_file_path = os.path.abspath(config_file_path)

# retrieves the name of the directory containing the configuration
# files and adds it to the system path, so that it's possible to
# import the configuration file module
configuration_directory_path = os.path.dirname(config_file_path)
if not configuration_directory_path in sys.path:
    sys.path.insert(0, configuration_directory_path)

# retrieves the configuration file base path from the configuration
# file path, this values is going to be used to retrieve the "final"
# module name to be imported in the python interpreter
config_file_base_path = os.path.basename(config_file_path)

# retrieves the configuration module name and the configuration
# module extension by splitting the configuration base path into
# base name and extension and then imports the referring module
configuration_module_name, _configuration_module_extension = os.path.splitext(config_file_base_path)
try: colony_configuration = __import__(configuration_module_name)
except ImportError: import colony.config.base as module; colony_configuration = module

# initializes the lists that will contain both the path to the
# plugins and the paths to the configuration (meta) files, these
# are only going to be used temporarily for glob expansion
_meta_paths = []
_plugin_paths = []

# iterates over each of the plugin paths to resolve them using
# the glob based approach then "takes" the final list into a
# final step of absolute path normalization
for plugin_path in colony_configuration.plugin_path_list:
    plugin_paths.append(os.path.join(manager_path, plugin_path))
for plugin_path in plugin_paths: _plugin_paths += glob.glob(plugin_path)
plugin_paths = [os.path.abspath(plugin_path) for plugin_path in _plugin_paths]

# iterates over each of the meta paths to resolve them using
# the glob based approach then "takes" the final list into a
# final step of absolute path normalization
for meta_path in colony_configuration.meta_path_list:
    meta_paths.append(os.path.join(manager_path, meta_path))
for meta_path in meta_paths: _meta_paths += glob.glob(meta_path)
meta_paths = [os.path.abspath(meta_path) for meta_path in _meta_paths]

# creates the plugin manager instance with the current file path
# as the manager path and the corresponding relative log path,
# then provides the plugin and meta paths and unsets the global
# loop strategy (avoids blocking the process), note also that
# threads are disallowed to avoid creation of extra threads, the
# signal handlers are disabled to avoid collisions
plugin_manager = colony.PluginManager(
    manager_path = manager_path,
    logger_path = os.path.join(manager_path, "log"),
    plugin_paths = plugin_paths,
    meta_paths = meta_paths,
    loop = False,
    threads = False,
    signals = False,
    layout_mode = layout_mode,
    run_mode = run_mode,
    container = "wsgi"
)
plugin_manager.start_logger(level)
return_code = plugin_manager.load_system()
alias = None
rewrite = None

def application(environ, start_response):
    try:
        # retrieves the currently set alias and rewrite lists,
        # loading them it in case this is the first run, these
        # value may be unset and no mapping is performed under
        # such situations (applicable for both alias and rewriting)
        alias = get_alias()
        rewrite = get_rewrite()

        # retrieves the wsgi plugin and uses it to handle
        # the wsgi request (request redirection) any inner
        # exception should be handled and an error HTTP
        # message should be returned to the end user
        wsgi_plugin = plugin_manager.get_plugin("pt.hive.colony.plugins.wsgi")
        if not wsgi_plugin: raise colony.PluginSystemException("no WSGI plugin found")
        sequence = wsgi_plugin.handle(environ, start_response, prefix, alias, rewrite)
    except Exception:
        # in case the run mode is development the exception should
        # be processed and a description sent to the output
        if run_mode == "development":
            # prints a description of the exception and then
            # sends the traceback of it to the output
            traceback.print_exc(file = sys.stdout)

        # raises the exception back to the stack to be handled
        # by the upper levels
        raise

    # returns the sequence object that may be used by the caller
    # method to retrieve the contents of the message to be sent
    return sequence

def get_alias(encoding = "utf-8"):
    global alias
    global alias_path

    # in case either the alias is defined (file already loaded)
    # or there's no alias file to be loaded the currently set
    # alias variable is retrieved (cached value)
    if alias or not alias_path: return alias

    try:
        # tries to load the JSON file using the default python
        # based JSON module (may not exist) the closes the file
        # to avoid any memory reference leak
        import json
        file = open(alias_path, "rb")
        try: data = file.read()
        finally: file.close()
        data = data.decode(encoding)
        alias = json.loads(data)
    except Exception:
        # unsets the alias path to avoid any further file reading
        # avoiding any duplicated reading
        alias_path = None

    # returns the map containing the alias that were loaded from the
    # file, this value may be unsets in case there was an error
    return alias

def get_rewrite(encoding = "utf-8"):
    global rewrite
    global rewrite_path

    # in case either the rewrite is defined (file already loaded)
    # or there's no rewrite file to be loaded the currently set
    # rewrite variable is retrieved (cached value)
    if rewrite or not rewrite_path: return rewrite

    try:
        # tries to load the JSON file using the default python
        # based JSON module (may not exist) the closes the file
        # to avoid any memory reference leak
        import json
        file = open(rewrite_path, "rb")
        try: data = file.read()
        finally: file.close()
        data = data.decode(encoding)
        rewrite = json.loads(data)
    except Exception:
        # unsets the rewrite path to avoid any further file reading
        # avoiding any duplicated reading
        rewrite_path = None

    # returns the map containing the rewrite that were loaded from the
    # file, this value may be unsets in case there was an error
    return rewrite

@atexit.register
def unload_system():
    # unloads the plugin manager system releasing all
    # the used resources and killing all the threads
    # this should be enough to return the control to
    # the embedding process
    plugin_manager.unload_system()

class ServerThread(threading.Thread):
    """
    Thread class responsible for the handling and management
    of the server instance that it represents.

    This is considered to be a long running thread and only
    the exit of the current process "kills" it.
    """

    def __init__(
        self,
        server = "netius",
        host = "127.0.0.1",
        port = 8080,
        ssl = False,
        key_file = None,
        cer_file = None,
        kwargs = dict(),
        *args,
        **_kwargs
    ):
        threading.Thread.__init__(self, *args, **_kwargs)
        self.server = server
        self.host = host
        self.port = port
        self.ssl = ssl
        self.key_file = key_file
        self.cer_file = cer_file
        self.kwargs = kwargs
        self.daemon = True

    def __repr__(self):
        return "%s / %s@%d" % (self.server, self.host, self.port)

    def run(self):
        try:
            serve(
                server = self.server,
                host = self.host,
                port = self.port,
                ssl = self.ssl,
                key_file = self.key_file,
                cer_file = self.cer_file,
                kwargs = self.kwargs
            )
        except Exception:
            sys.stderr.write("Problem in '%s'" % str(self) + "\n")
            raise

def serve_multiple(
    server = "netius",
    hosts = ("127.0.0.1",),
    ports = (8080,),
    ssl = False,
    key_file = None,
    cer_file = None,
    kwargs = dict()
):
    threads = []
    count = len(hosts)

    for index in range(count):
        host = hosts[index]
        port = ports[index]

        thread = ServerThread(
            server = server,
            host = host,
            port = port,
            ssl = ssl,
            key_file = key_file,
            cer_file = cer_file,
            kwargs = kwargs
        )
        thread.start()
        threads.append(thread)

    return threads

def serve(
    server = "netius",
    host = "127.0.0.1",
    port = 8080,
    ssl = False,
    key_file = None,
    cer_file = None,
    kwargs = dict()
):
    _globals = globals()
    sys.stderr.write("Starting with '%s' ..." % server + "\n")
    method = _globals.get("serve_" + server, serve_legacy)
    return_value = method(
        host = host,
        port = port,
        ssl = ssl,
        key_file = key_file,
        cer_file = cer_file,
        **kwargs
    )
    sys.stderr.write("Stopped in '%s' ..." % server + "\n")
    return return_value

def serve_waitress(host, port, **kwargs):
    import waitress
    waitress.serve(application, host = host, port = port)

def serve_netius(
    host,
    port,
    ssl = False,
    key_file = None,
    cer_file = None,
    **kwargs
):
    import netius.servers
    server = netius.servers.WSGIServer(application, **kwargs)
    server.serve(
        host = host,
        port = port,
        ssl = ssl,
        key_file = key_file,
        cer_file = cer_file
    )

def serve_tornado(
    host,
    port,
    ssl = False,
    key_file = None,
    cer_file = None,
    **kwargs
):
    import tornado.wsgi
    import tornado.httpserver

    ssl_options = ssl and dict(
        keyfile = key_file,
        certfile = cer_file
    ) or None

    container = tornado.wsgi.WSGIContainer(application)
    server = tornado.httpserver.HTTPServer(container, ssl_options = ssl_options)
    server.listen(port, address = host)
    instance = tornado.ioloop.IOLoop.instance()
    instance.start()

def serve_cherry(host, port, **kwargs):
    import cherrypy.wsgiserver

    server = cherrypy.wsgiserver.CherryPyWSGIServer(
        (host, port),
        application
    )
    try: server.start()
    except (KeyboardInterrupt, SystemExit): server.stop()

def serve_legacy(host, port, **kwargs):
    import wsgiref.simple_server
    httpd = wsgiref.simple_server.make_server(host, port, application)
    sys.stderr.write("Running on http://%s:%d/" % (host, port) + "\n")
    httpd.serve_forever()

def main():
    kwargs = dict()
    server = colony.conf("SERVER", "legacy")
    host = colony.conf("HOST", "127.0.0.1")
    port = colony.conf("PORT", "8080")
    ssl = colony.conf("SSL", False, cast = bool)
    key_file = colony.conf("KEY_FILE", None)
    cer_file = colony.conf("CER_FILE", None)
    for name, value in os.environ.items():
        if not name.startswith("SERVER_"): continue
        name_s = name.lower()[7:]
        if name_s in EXCLUDED_NAMES: continue
        kwargs[name_s] = value

    host = str(host)
    port = str(port)

    hosts = [value.strip() for value in host.split(",")]
    ports = [int(value.strip()) for value in port.split(",")]

    serve_multiple(
        server = server,
        hosts = hosts,
        ports = ports,
        ssl = ssl,
        key_file = key_file,
        cer_file = cer_file,
        kwargs = kwargs
    )

    def handler(signum = None, frame = None): raise SystemExit()
    signal.signal(signal.SIGTERM, handler)

    run = True
    while run:
        try: time.sleep(86400)
        except (KeyboardInterrupt, SystemExit): run = False

if __name__ == "__main__":
    main()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import json

from . import legacy

FILE_NAME = "colony.json"
""" The default name of the file that is going to be
used for the loading of configuration values from JSON """

FILE_TEMPLATE = "colony.%s.json"
""" The template to be used in the construction of the
domain specific configuration file paths """

HOME_FILE = "~/.home"
""" The location of the file that may be used to "redirect"
the home directory contents to a different directory """

IMPORT_NAMES = ("$import", "$include", "$IMPORT", "$INCLUDE")
""" The multiple possible definitions of the special configuration
name that references a list of include files to be loaded """

CASTS = {
    bool : lambda v: v if isinstance(v, bool) else v in ("1", "true", "True"),
    list : lambda v: v if isinstance(v, list) else v.split(";") if v else [],
    tuple : lambda v: v if isinstance(v, tuple) else tuple(v.split(";") if v else [])
}
""" The map containing the various cast method
operation associated with the various data types,
they provide a different type of casting strategy """

ENV_ENCODINGS = (
    "utf-8",
    sys.getdefaultencoding(),
    sys.getfilesystemencoding()
)
""" The sequence of encodings that are going to
be used to try to decode possible byte based strings
for the various environment variable values """

CONFIGS = {}
""" The map that contains the key value association
for all the currently set global configurations """

CONFIG_F = []
""" The list of files that have been used for the loading
of the configuration through this session, every time a
loading of configuration from a file occurs the same path
is added to this global list """

HOMES = []
""" Global reference to the paths to the directory considered
to be the home on in terms of configuration, this value should
be set on the initial loading of the ".home" file """

GLOBAL_CONFIG = {
    "logging_format" : "%(asctime)s [%(levelname)s] %(message)s",
    "plugin_id_logging" : True,
    "thread_id_logging" : True
}
""" The plugin manager configuration, this value is
used only as part of the inner (private) configuration
of the colony infra-structure and should only be changed
with complete knowledge of the inner workings of the system """

if not isinstance(__builtins__, dict):
    __builtins__ = __builtins__.__dict__

def conf(name, default = None, cast = None, ctx = None):
    """
    Retrieves the configuration value for the provided value
    defaulting to the provided default value in case no value
    is found for the provided name.

    An optional cast operation may be performed on the value
    in case it's requested.

    :type name: String
    :param name: The name of the configuration value to be
    retrieved.
    :type default: Object
    :param default: The default value to be retrieved in case
    no value was found for the provided name.
    :type cast: Type/String
    :param cast: The cast operation to be performed in the
    resolved value (optional).
    :type ctx: Dictionary
    :param ctx: The context dictionary to be used for situations
    where a more contextual configuration is meant to be used instead
    of the process wide global configuration.
    :rtype: Object
    :return: The value for the configuration with the requested
    name or the default value if no value was found.
    """

    configs = ctx["configs"] if ctx else CONFIGS
    cast = _cast_r(cast)
    value = configs.get(name, default)
    if cast and not value == None: value = cast(value)
    return value

def conf_prefix(prefix, ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    configs_prefix = dict()
    for name, value in configs.items():
        if not name.startswith(prefix): continue
        configs_prefix[name] = value
    return configs_prefix

def conf_suffix(suffix, ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    configs_suffix = dict()
    for name, value in configs.items():
        if not name.endswith(suffix): continue
        configs_suffix[name] = value
    return configs_suffix

def conf_s(name, value, ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    configs[name] = value

def conf_r(name, ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    if not name in configs: return
    del configs[name]

def conf_d(ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    return configs

def conf_ctx():
    return dict(configs = dict(), config_f = dict())

def load(names = (FILE_NAME,), path = None, encoding = "utf-8", ctx = None):
    paths = []
    homes = get_homes()
    for home in homes:
        paths += [
            os.path.join(home),
            os.path.join(home, ".config"),
        ]
    paths += [sys.prefix]
    paths.append(path)
    for path in paths:
        for name in names:
            load_file(name = name, path = path, encoding = encoding, ctx = ctx)
    load_env(ctx = ctx)

def load_file(name = FILE_NAME, path = None, encoding = "utf-8", ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS
    config_f = ctx["config_f"] if ctx else CONFIG_F

    if path: path = os.path.normpath(path)
    if path: file_path = os.path.join(path, name)
    else: file_path = name

    file_path = os.path.abspath(file_path)
    file_path = os.path.normpath(file_path)
    base_path = os.path.dirname(file_path)

    exists = os.path.exists(file_path)
    if not exists: return

    exists = file_path in config_f
    if exists: config_f.remove(file_path)
    config_f.append(file_path)

    file = open(file_path, "rb")
    try: data = file.read()
    finally: file.close()
    if not data: return

    data = data.decode(encoding)
    data_j = json.loads(data)

    _load_includes(base_path, data_j, encoding = encoding)

    for key, value in data_j.items():
        if not _is_valid(key): continue
        configs[key] = value

def load_env(ctx = None):
    configs = ctx["configs"] if ctx else CONFIGS

    config = dict(os.environ)
    homes = get_homes()

    for home in homes:
        _load_includes(home, config)

    for key, value in legacy.iteritems(config):
        if not _is_valid(key): continue
        configs[key] = value
        is_bytes = legacy.is_bytes(value)
        if not is_bytes: continue
        for encoding in ENV_ENCODINGS:
            try: value = value.decode(encoding)
            except UnicodeDecodeError: pass
            else: break
        configs[key] = value

def get_homes(
    file_path = HOME_FILE,
    default = "~",
    encoding = "utf-8",
    force_default = False
):
    global HOMES
    if HOMES: return HOMES

    HOMES = os.environ.get("HOMES", None)
    HOMES = HOMES.split(";") if HOMES else HOMES
    if not HOMES == None: return HOMES

    default = os.path.expanduser(default)
    default = os.path.abspath(default)
    default = os.path.normpath(default)
    HOMES = [default]

    file_path = os.path.expanduser(file_path)
    file_path = os.path.normpath(file_path)
    exists = os.path.exists(file_path)
    if not exists: return HOMES

    if not force_default: del HOMES[:]

    file = open(file_path, "rb")
    try: data = file.read()
    finally: file.close()

    data = data.decode("utf-8")
    data = data.strip()
    paths = data.splitlines()
    paths = [path.strip() for path in paths]

    for path in paths:
        path = path.strip()
        if not path: continue
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        path = os.path.normpath(path)
        HOMES.append(path)

    return HOMES

def _cast_r(cast):
    is_string = type(cast) in legacy.STRINGS
    if is_string: cast = __builtins__.get(cast, None)
    if not cast: return None
    return CASTS.get(cast, cast)

def _load_includes(base_path, config, encoding = "utf-8"):
    includes = ()

    for alias in IMPORT_NAMES:
        includes = config.get(alias, includes)

    if legacy.is_string(includes):
        includes = includes.split(";")

    for include in includes:
        load_file(
            name = include,
            path = base_path,
            encoding = encoding
        )

def _is_valid(key):
    if key in IMPORT_NAMES: return False
    return True

load()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import threading

class UpdateThread(threading.Thread):
    """
    The update thread class.
    """

    stop_flag = False
    """ The stop flag """

    timeout = None
    """ The timeout between calls """

    call_method = None
    """ The method to be called """

    call_arguments = []
    """ The call arguments """

    def __init__(self):
        """
        Constructor of the class.
        """

        threading.Thread(self)

        self.daemon = True

    def stop(self):
        """
        Stops the update thread.
        """

        self.stop_flag = True

    def get_timeout(self):
        """
        Retrieves the timeout value.

        :rtype: int
        :return: The timeout value.
        """

        return self.timeout

    def set_timeout(self, timeout):
        """
        Sets the timeout value.

        :type timeout: int
        :param timeout: The timeout value.
        """

        self.timeout = timeout

    def get_call_method(self):
        """
        Retrieves the call method.

        :rtype: Method
        :return: The call method.
        """

        return self.call_method

    def set_call_method(self, call_method):
        """
        Sets the call method.

        :type call_method: Method
        :param call_method: The call method.
        """

        self.call_method = call_method

    def get_call_arguments(self):
        """
        Retrieves the call arguments.

        :rtype: List
        :return: The call arguments.
        """

        return self.call_arguments

    def set_call_arguments(self, call_arguments):
        """
        Sets the call arguments.

        :type call_arguments: List
        :param call_arguments: The call argument.
        """

        self.call_arguments = call_arguments

    def run(self):
        # unsets the stop flag
        self.stop_flag = False

        # while the thread is valid
        while not self.stop_flag:
            # sleep for the given timeout time
            time.sleep(self.timeout)

            # in case the stop flag is set
            if self.stop_flag:
                # breaks the cycle
                break

            # calls the method
            self.call_method(*self.call_arguments)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import colony

class QuoteTest(colony.ColonyTestCase):
    """
    Class that tests the various functions/methods related
    with the quoting/unquoting process of strings.
    """

    def test_quote(self):
        """
        Validates/verifies that the quoting support is complete
        and correctly working. Uses a lot of languages to try
        to add complexity to the set of tests.
        """

        result = colony.quote("Hello World")
        self.assertEqual(result, "Hello%20World")

        result = colony.quote("Olá Mundo")
        self.assertEqual(result, "Ol%C3%A1%20Mundo")

        result = colony.quote("你好世界")
        self.assertEqual(result, "%E4%BD%A0%E5%A5%BD%E4%B8%96%E7%95%8C")

    def test_quote_plus(self):
        """
        Plus version testing of the quoting operation, this test is
        analogous to the previous test.
        """

        result = colony.quote_plus("Hello World")
        self.assertEqual(result, "Hello+World")

        result = colony.quote_plus("Olá Mundo")
        self.assertEqual(result, "Ol%C3%A1+Mundo")

        result = colony.quote_plus("你好世界")
        self.assertEqual(result, "%E4%BD%A0%E5%A5%BD%E4%B8%96%E7%95%8C")

    def test_unquote(self):
        """
        Validates and verifies that the unquoting (reverse operation)
        works using the default infra-structure. Will try to decode
        results from a variety of languages.
        """

        result = colony.unquote("Hello%20World")
        self.assertEqual(result, "Hello World")

        result = colony.unquote("Ol%C3%A1%20Mundo")
        self.assertEqual(result, "Olá Mundo")

        result = colony.unquote("%E4%BD%A0%E5%A5%BD%E4%B8%96%E7%95%8C")
        self.assertEqual(result, "你好世界")

        result = colony.unquote("Hello%20World%GG")
        self.assertEqual(result, "Hello World%GG")

        result = colony.unquote("你好世界", strict = False)
        self.assertEqual(result, "你好世界")

        result = colony.unquote(b"\xe4\xbd\xa0\xe5\xa5\xbd\xe4\xb8\x96\xe7\x95\x8c", strict = False)
        self.assertEqual(result, "你好世界")

        result = colony.unquote(colony.legacy.u("你好世界"), strict = False)
        self.assertEqual(result, "你好世界")

        self.assert_raises(
            UnicodeDecodeError,
            colony.unquote,
            b"\xe4\xbd\xa0\xe5\xa5\xbd\xe4\xb8\x96\xe7\x95\x8c",
            strict = True
        )

        self.assert_raises(
            UnicodeEncodeError,
            colony.unquote,
            colony.legacy.u("你好世界"),
            strict = True
        )

    def test_unquote_plus(self):
        """
        Plus version testing of the unquoting operation, this test is
        analogous to the previous test.
        """

        result = colony.unquote_plus("Hello+World")
        self.assertEqual(result, "Hello World")

        result = colony.unquote_plus("Ol%C3%A1+Mundo")
        self.assertEqual(result, "Olá Mundo")

        result = colony.unquote_plus("%E4%BD%A0%E5%A5%BD%E4%B8%96%E7%95%8C")
        self.assertEqual(result, "你好世界")

        result = colony.unquote_plus("Hello+World%GG")
        self.assertEqual(result, "Hello World%GG")

        result = colony.unquote_plus("你好世界", strict = False)
        self.assertEqual(result, "你好世界")

        result = colony.unquote_plus(b"\xe4\xbd\xa0\xe5\xa5\xbd\xe4\xb8\x96\xe7\x95\x8c", strict = False)
        self.assertEqual(result, "你好世界")

        result = colony.unquote_plus(colony.legacy.u("你好世界"), strict = False)
        self.assertEqual(result, "你好世界")

        self.assert_raises(
            UnicodeDecodeError,
            colony.unquote_plus,
            b"\xe4\xbd\xa0\xe5\xa5\xbd\xe4\xb8\x96\xe7\x95\x8c",
            strict = True
        )

        self.assert_raises(
            UnicodeEncodeError,
            colony.unquote_plus,
            colony.legacy.u("你好世界"),
            strict = True
        )

    def test_url_encode(self):
        """
        Test for the URL encode operation that encodes the key to value
        attributes that is present in the query string for URLs.
        """

        items = (("message", "Hello World"), ("mensagem", "Olá Mundo"))

        result = colony.url_encode(attributes_list = items, plus_encoding = False)
        self.assertEqual(result, "message=Hello%20World&mensagem=Ol%C3%A1%20Mundo")

        result = colony.url_encode(attributes_list = items, plus_encoding = True)
        self.assertEqual(result, "message=Hello+World&mensagem=Ol%C3%A1+Mundo")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time

import colony

class SchedulerTest(colony.ColonyTestCase):
    """
    Class that tests the scheduler up to the expected
    values, making sure that race and other weird conditions
    are properly handled.
    """

    def setUp(self):
        self.scheduler = colony.Scheduler()
        self.scheduler.start_scheduler()

    def tearDown(self):
        self.scheduler.stop_scheduler()
        self.scheduler.join(10)

    def test_basic(self):
        """
        Runs a series of basic sanity tests for the scheduler.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        values = dict()

        def update_values():
            values["a"] = 1

        self.assertEqual(values, dict())

        identifier = self.scheduler.add_callable(update_values)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(identifier, 1)
        self.assertEqual(values, dict(a = 1))
        self.assertEqual(self.scheduler.is_running(), True)
        self.assertEqual(self.scheduler.is_busy(), False)

        self.scheduler.reset_scheduler()
        self.assertEqual(self.scheduler.is_running(), False)
        self.assertEqual(self.scheduler.is_busy(), False)

    def test_delayed(self):
        """
        Tests that a delayed work can be scheduled and executed
        in the proper time.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        values = dict()

        def update_values():
            values["a"] = 1

        self.assertEqual(values, dict())

        initial = time.time()
        identifier = self.scheduler.add_callable(update_values, timestamp = time.time() + 0.3)
        self.assertEqual(identifier, 1)
        time.sleep(0.1)
        self.assertEqual(values, dict())

        self.scheduler.wait_callable(identifier)
        self.assertEqual(time.time() - initial >= 0.3, True)
        self.assertEqual(values, dict(a = 1))

    def test_multiple(self):
        """
        Tests that multiple parallel callable tasks can be
        scheduled and executed.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        values = dict()

        def update_values_1():
            values["a"] = 1

        def update_values_2():
            values["b"] = 2

        self.assertEqual(values, dict())

        identifier_1 = self.scheduler.add_callable(update_values_1)
        identifier_2 = self.scheduler.add_callable(update_values_2)
        self.scheduler.wait_callable(identifier_1)
        self.scheduler.wait_callable(identifier_2)
        self.assertEqual(identifier_1, 1)
        self.assertEqual(identifier_2, 2)
        self.assertEqual(values, dict(a = 1, b = 2))
        self.assertEqual(self.scheduler.is_running(), True)
        self.assertEqual(self.scheduler.is_busy(), False)

        values = dict()

        def update_values_3():
            values["a"] = 1

        def update_values_4():
            values["b"] = values["a"] + 1

        self.assertEqual(values, dict())

        identifier_1 = self.scheduler.add_callable(update_values_3, timestamp = time.time() + 0.1)
        identifier_2 = self.scheduler.add_callable(update_values_4, timestamp = time.time() + 0.2)
        self.scheduler.wait_callable(identifier_1)
        self.scheduler.wait_callable(identifier_2)
        self.assertEqual(identifier_1, 3)
        self.assertEqual(identifier_2, 4)
        self.assertEqual(values, dict(a = 1, b = 2))
        self.assertEqual(self.scheduler.is_running(), True)
        self.assertEqual(self.scheduler.is_busy(), False)

        values = dict()

        def update_values_5():
            values["a"] = values["b"] + 1

        def update_values_6():
            values["b"] = 2

        self.assertEqual(values, dict())

        identifier_1 = self.scheduler.add_callable(update_values_5, timestamp = time.time() + 0.2)
        identifier_2 = self.scheduler.add_callable(update_values_6, timestamp = time.time() + 0.1)
        self.scheduler.wait_callable(identifier_1)
        self.scheduler.wait_callable(identifier_2)
        self.assertEqual(identifier_1, 5)
        self.assertEqual(identifier_2, 6)
        self.assertEqual(values, dict(a = 3, b = 2))
        self.assertEqual(self.scheduler.is_running(), True)
        self.assertEqual(self.scheduler.is_busy(), False)

    def test_overflow(self):
        """
        Tests that it is possible to add a task that overflows
        the max timeout value for the scheduler.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        values = dict()

        def update_values():
            values["a"] = 1

        self.assertEqual(values, dict())

        identifier = self.scheduler.add_callable(
            update_values,
            timestamp = time.time() + colony.SCHEDULING_MAX + 1
        )
        self.assertEqual(identifier, 1)
        self.assertEqual(values, dict())
        self.assertEqual(self.scheduler.is_running(), True)

        update_values()
        self.assertEqual(values, dict(a = 1))

    def test_stopped(self):
        """
        Tests that if the scheduler is currently stopped then
        a series of consequences must happen.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        self.scheduler.stop_scheduler()
        self.assertEqual(self.scheduler.join(10), None)
        self.assertEqual(self.scheduler.is_running(), False)
        self.assert_raises(colony.AssertionError, lambda: self.scheduler.add_callable(lambda: 1))

        values = dict()

        def update_values():
            values["a"] = 1

        self.assertEqual(values, dict())

        identifier = self.scheduler.add_callable(update_values, verify = False)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(values, dict())

        self.assert_raises(RuntimeError, self.scheduler.start_scheduler)

        update_values()
        self.assertEqual(values, dict(a = 1))

    def test_exception_handler(self):
        """
        Tests that the exception handler is properly handling
        exceptions raised in the callable execution.
        """

        self.assertEqual(self.scheduler.is_running(), True)
        self.assertEqual(self.scheduler.exception_handler, None)

        values = dict()

        def update_values_raise():
            values["a"] = 1
            raise Exception()

        identifier = self.scheduler.add_callable(update_values_raise)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(values, dict(a = 1))
        self.assertEqual(self.scheduler.is_running(pedantic = True), True)

        def exception_handler(callable, exception):
            values["callable"] = callable
            values["exception"] = exception.__class__

        self.scheduler.set_exception_handler(exception_handler)

        identifier = self.scheduler.add_callable(update_values_raise)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(
            values,
            dict(
                a = 1,
                callable = update_values_raise,
                exception = Exception
            )
        )
        self.assertEqual(self.scheduler.is_running(pedantic = True), True)

    def test_waiting(self):
        """
        Tests the waiting operation, specially for edge cases.
        """

        self.assertEqual(self.scheduler.is_running(), True)

        values = dict()

        def update_values_1():
            values["a"] = 1

        self.assertEqual(values, dict())

        identifier = self.scheduler.add_callable(update_values_1)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(identifier, 1)
        self.assertEqual(values, dict(a = 1))

        values = dict()

        def update_values_2():
            values["a"] = 2

        self.assertEqual(values, dict())

        identifier = self.scheduler.add_callable(update_values_2)
        time.sleep(0.1)
        self.scheduler.wait_callable(identifier)
        self.assertEqual(identifier, 2)
        self.assertEqual(values, dict(a = 2))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import xml.dom.minidom

import colony

class XmlTest(colony.ColonyTestCase):
    """
    Test class for the complete set of functions associated
    with XML manipulation.
    """

    def test_xml_to_dict(self):
        """
        Verifies a series of conditions associated with XML to
        dictionary conversion.
        """

        result = colony.xml_to_dict("""<person>
            <name>Hello World</name>
            <age>32</age>
        </person>""")
        self.assertEqual(
            result,
            dict(
                person = dict(
                    name = "Hello World",
                    age = "32"
                )
            )
        )

        result = colony.xml_to_dict("""<person>
            <name>Hello World</name>
            <age>32</age>
            <address>
                <street>Wood Street</street>
                <city>London</city>
                <country>United Kingdom</country>
            </address>
        </person>""")
        self.assertEqual(
            result,
            dict(
                person = dict(
                    name = "Hello World",
                    age = "32",
                    address = dict(
                        street = "Wood Street",
                        city = "London",
                        country = "United Kingdom"
                    )
                )
            )
        )

        result = colony.xml_to_dict("""<person>
            <name>Hello World</name>
            <age>32</age>
            <address></address>
        </person>""")
        self.assertEqual(
            result,
            dict(
                person = dict(
                    name = "Hello World",
                    age = "32",
                    address = None
                )
            )
        )

        result = colony.xml_to_dict("""<person>
            <name>你好世界</name>
            <age>32</age>
        </person>""")
        self.assertEqual(
            result,
            dict(
                person = dict(
                    name = colony.legacy.u("你好世界"),
                    age = "32"
                )
            )
        )

        result = colony.xml_to_dict(
            xml.dom.minidom.parseString("""<person>
                <name>Hello World</name>
                <age>32</age>
            </person>""")
        )
        self.assertEqual(
            result,
            dict(
                person = dict(
                    name = "Hello World",
                    age = "32"
                )
            )
        )

    def test_dict_to_xml(self):
        """
        Verifies a series of conditions associated with dictionary
        to XML conversion
        """

        result = colony.dict_to_xml(dict(
            person = dict(
                name = "Hello World",
                age = "32"
            )
        ))
        self.assertEqual(result, "<person><age>32</age><name>Hello World</name></person>")

        result = colony.dict_to_xml(dict(
            person = dict(
                name = "Hello World",
                age = "32",
                address = dict(
                    street = "Wood Street",
                    city = "London",
                    country = "United Kingdom"
                )
            )
        ))
        self.assertEqual(result, "<person><address><city>London</city><country>United Kingdom</country><street>Wood Street</street></address><age>32</age><name>Hello World</name></person>")

        result = colony.dict_to_xml(dict(
            person = dict(
                name = "Hello World",
                age = "32",
                address = None
            )
        ))
        self.assertEqual(result, "<person><address></address><age>32</age><name>Hello World</name></person>")

        result = colony.dict_to_xml(dict(
            person = dict(
                name = "你好世界",
                age = "32"
            )
        ))
        self.assertEqual(result, colony.legacy.u("<person><age>32</age><name>你好世界</name></person>"))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.3"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import stat

BUFFER_SIZE = 4096
""" The size of the buffer for file operations """

LONG_PATH_PREFIX = "\\\\?\\"
""" The windows long path prefix """

CURRENT_DIRECTORY = "."
""" The current directory string """

PARENT_DIRECTORY = ".."
""" The parent directory string """

SEPARATOR = "/"
""" The directory separator string """

NT_PLATFORM_VALUE = "nt"
""" The nt platform value """

CE_PLATFORM_VALUE = "ce"
""" The ce platform value """

DOS_PLATFORM_VALUE = "dos"
""" The dos platform value """

WINDOWS_PLATFORMS_VALUE = (
    NT_PLATFORM_VALUE,
    CE_PLATFORM_VALUE,
    DOS_PLATFORM_VALUE
)
""" The windows platform value """

def normalize_path(path):
    """
    Normalizes the given path, using the characteristics
    of the current environment.
    In windows this function adds support for long path names.

    :type path: String
    :param path: The path to be normalized.
    :rtype: String
    :return: The normalized path.
    """

    # retrieves the current os name
    os_name = os.name

    # in case the current operative system is windows based and
    # the normalized path does start with the long path prefix it
    # must be removed to allow a "normal" path normalization
    if os_name in WINDOWS_PLATFORMS_VALUE and path.startswith(LONG_PATH_PREFIX):
        # removes the long path prefix from the path
        path = path[4:]

    # checks if the path is absolute
    is_absolute_path = os.path.isabs(path)

    # in case the path is not absolute (creates problem in windows
    # long path support)
    if os_name in WINDOWS_PLATFORMS_VALUE and not is_absolute_path:
        # converts the path to absolute
        path = os.path.abspath(path)

    # normalizes the path
    normalized_path = os.path.normpath(path)

    # in case the current operative system is windows based and
    # the normalized path does not start with the long path prefix
    if os_name in WINDOWS_PLATFORMS_VALUE and not normalized_path.startswith(LONG_PATH_PREFIX):
        # creates the path in the windows mode, adds
        # the support for long path names with the prefix token
        normalized_path = LONG_PATH_PREFIX + normalized_path

    # returns the normalized path
    return normalized_path

def align_path(path):
    """
    Aligns the given path, converting all the system specific
    characters into the defined virtual separators.
    The retrieved path is system independent.

    :type path: String
    :param path: The path to the aligned (become system independent).
    :rtype: String
    :return: The aligned path (system independent).
    """

    # aligns the path replacing the backslashes with
    # "normal" slashes
    aligned_path = path.replace("\\", "/")

    # returns the aligned path
    return aligned_path

def copy_directory(source_path, target_path, replace_files = True, copy_hidden = True):
    """
    Copies the directory in the given source path to the
    target path.
    The copy is recursive and so the sub directories are copied too.
    The directory in the target path is created if not existent
    or overwritten if existent.
    It's possible to control the copying of hidden files using the
    optional copy hidden flag parameter.

    :type source_path: String
    :param source_path: The path to the source directory.
    :type target_path: String
    :param target_path: The path to the target directory.
    :type replace_files: bool
    :param replace_files: If the files should be replaced
    in case duplicate files are found.
    :type copy_hidden: bool
    :param copy_hidden: If the files considered by the os to
    be of type hidden should be copied..
    """

    # normalizes both the target and source paths
    # (avoids path problems in various platforms)
    source_path = normalize_path(source_path)
    target_path = normalize_path(target_path)

    # in case the source path is not a directory
    if not os.path.isdir(source_path):
        # raises an exception
        raise Exception("Source path is not a directory: '%s" % source_path)

    # in case the target path does not exist
    if not os.path.exists(target_path):
        # creates the target path directory
        os.makedirs(target_path)

    # in case the target path is not a directory
    if not os.path.isdir(target_path):
        # raises an exception
        raise Exception("Target path is not a directory: '%s" % target_path)

    # retrieves the directory list from the source path
    directory_list = os.listdir(source_path)

    # iterates over all the entry names in the
    # directory list
    for entry_name in directory_list:
        # creates the entry full path from the source path
        # and the entry name and normalizes it
        entry_full_path = os.path.join(source_path, entry_name)
        entry_full_path = normalize_path(entry_full_path)

        # creates the target full path and normalizes it
        target_full_path = os.path.join(target_path, entry_name)
        target_full_path = normalize_path(target_full_path)

        # retrieve the base name for the current entry
        # for hidden checking
        entry_base_name = os.path.basename(entry_full_path)

        # in case the copy hidden flag is not set and the base
        # name refers a hidden file or directory
        if not copy_hidden and entry_base_name.startswith("."):
            # continues the loop (no copy)
            continue

        # retrieves the mode
        mode = os.lstat(entry_full_path)[stat.ST_MODE]

        # in case it is a directory
        if stat.S_ISDIR(mode):
            # copies the (sub) directory
            copy_directory(entry_full_path, target_full_path, replace_files, copy_hidden)
        # in case it is a symbolic link (special
        # care must be taken in such case)
        elif stat.S_ISLNK(mode):
            # copies the symbolic link to the target path
            copy_link(entry_full_path, target_full_path, replace_files)
        # otherwise it's a file and must be copied
        else:
            # copies the entry to the target path
            copy_file(entry_full_path, target_full_path, replace_files)

def copy_link(source_path, target_path, replace_file = True):
    """
    Copies a symbolic link in the given source path to the
    target path.
    The symbolic link in the target path is created if not existent
    or overwritten if existent.

    :type source_path: String
    :param source_path: The path to the source symbolic link.
    :type target_path: String
    :param target_path: The path to the target symbolic link.
    :type replace_file: bool
    :param replace_file: If the file should be replaced
    in existent symbolic link is found.
    """

    # checks if the target path (symbolic link) exists
    target_file_exists = os.path.exists(target_path)

    # in case the replace file flag is not set and the
    # target path (symbolic link) exists (avoids
    # replacing symbolic link)
    if not replace_file and target_file_exists:
        # returns immediately (no copy)
        return

    # reads the link target (path) from the source path
    # and then uses that value to create the link in the
    # target path (link copy)
    link_target = os.readlink(source_path) #@UndefinedVariable
    os.symlink(link_target, target_path) #@UndefinedVariable

def copy_file(source_path, target_path, replace_file = True):
    """
    Copies a file in the given source path to the
    target path.
    The file in the target path is created if not existent
    or overwritten if existent.

    :type source_path: String
    :param source_path: The path to the source file.
    :type target_path: String
    :param target_path: The path to the target file.
    :type replace_file: bool
    :param replace_file: If the file should be replaced
    in existent files is found.
    """

    # normalizes both the target and source paths
    # (avoids path problems in various platforms)
    source_path = normalize_path(source_path)
    target_path = normalize_path(target_path)

    # checks if the target path (file) exists
    target_file_exists = os.path.exists(target_path)

    # in case the replace file flag is not set and the
    # target path (file) exists (avoids replacing file)
    if not replace_file and target_file_exists:
        # returns immediately (no copy)
        return

    # opens the source file
    source_file = open(source_path, "rb")

    # opens the target file
    target_file = open(target_path, "wb")

    try:
        # iterates continuously
        while True:
            # reads contents from the source file
            contents = source_file.read(BUFFER_SIZE)

            # in case the contents are not valid
            if not contents:
                # breaks the cycle
                break

            # writes the contents in the target file
            target_file.write(contents)
    finally:
        # closes the source file
        source_file.close()

        # closes the target file
        target_file.close()

def remove_directory(directory_path):
    """
    Removes the given directory path recursively.
    Directories containing files will have their contents removed
    before being removed.

    :type directory_path: String
    :param directory_path: The path to the directory to be removed.
    """

    # normalizes the directory path
    # (avoids path problems in various platforms)
    directory_path = normalize_path(directory_path)

    # creates the list of paths for the directory path
    paths_list = [os.path.join(directory_path, file_path) for file_path in os.listdir(directory_path)]

    # iterates over all the paths in the paths
    # list to remove them
    for path in paths_list:
        # normalizes the path (avoids
        # problems in various platforms)
        path = normalize_path(path)

        # in case the path is a directory
        if os.path.isdir(path):
            # removes the directory
            remove_directory(path)
        # otherwise it must be a "normal" file
        else:
            # removes the path
            os.remove(path)

    # removes the directory
    os.rmdir(directory_path)

def link(target_path, link_path, link_name = True, replace = False):
    """
    Creates a link between the target path and the link
    path given.
    In case the link name flag is set the link is relative
    and only the base name of the link path is used, this
    should provide extra flexibility to the link.

    An optional replace flag may be set so that the link is
    removed in case it already exists.

    :type target_path: String
    :param target_path: The target path to the link.
    :type link_path: String
    :param link_path: The path to the link.
    :type link_name: bool
    :param link_name: If the link base name should be used
    instead of the full path.
    :type replace: bool
    :param replace: If the link path should be replace (overwritten)
    in case it already exists (file overlapping).
    """

    # in case there is an existent link in the link path
    # it must be removed (the link may be a directory), this
    # behavior is only accomplished when the replace flag
    # parameter is set
    if replace and os.path.islink(link_path): os.remove(link_path)
    elif replace and os.path.isdir(link_path): remove_directory(link_path)

    # in case the current platform is windows
    if os.name == NT_PLATFORM_VALUE:
        # copies the directory (symbolic link through
        # the copy method)
        link_copy(target_path, link_path)
    # otherwise the platform probably supports the
    # normal file linking system (symbolic link)
    else:
        # retrieves the base name of the target path
        # as the target path name
        target_path_name = os.path.basename(target_path)

        # selects the correct target path based in the link name flag
        target_path = link_name and target_path_name or target_path

        # creates the symbolic link to the target path
        # using the link path
        os.symlink(target_path, link_path) #@UndefinedVariable

def link_copy(target_path, link_path):
    """
    Copies the target path into the path defined
    in the link path.
    This function acts as a stub for the symlink function.

    :type target_path: String
    :param target_path: The target path to the link.
    :type link_path: String
    :param link_path: The path to the "link".
    """

    # in case the directory exists
    if os.path.exists(link_path):
        # removes the directory in the link path
        remove_directory(link_path)

    # in case the target path reference a directory
    # a directory copy must be done
    if os.path.isdir(target_path):
        # copies the directory in the target path to the link path
        copy_directory(target_path, link_path)
    # otherwise a "simple" file copy is done
    else:
        # copies the file in the target path to the link path
        copy_file(target_path, link_path)

def ensure_file_path(file_path, default_file_path):
    """
    Ensures that the given file path is set with
    contents.
    In case the file does not exists the file in the default
    file path is copied to the file path.

    :type file_path: String
    :param file_path: The file path to ensure contents.
    :type default_file_path: String
    :param default_file_path: The path to the file to be
    used in case the file path does not exist.
    """

    # normalizes both the base and default
    # file paths (avoids path problems in
    # various platforms)
    file_path = normalize_path(file_path)
    default_file_path = normalize_path(default_file_path)

    # checks if the file exists
    file_exists = os.path.exists(file_path)

    # in case the file already exists
    if file_exists:
        # returns immediately
        return

    # retrieves the file directory path
    file_directory_path = os.path.dirname(file_path)

    # checks if the file directory path exists
    file_directory_path_exists = os.path.isdir(file_directory_path)

    # in case the file directory path does not exists creates the
    # directories required recursively
    not file_directory_path_exists and os.makedirs(file_directory_path)

    # opens the default file
    default_file = open(default_file_path, "rb")

    try:
        # reads the default file contents
        default_file_contents = default_file.read()
    finally:
        # closes the default file
        default_file.close()

    # opens the file
    file = open(file_path, "wb")

    try:
        # writes the default file contents to it
        # in order to set the value
        file.write(default_file_contents)
    finally:
        # closes the file
        file.close()

def is_parent_path(path, parent_path):
    """
    Checks if the given parent path is a parent
    to the given path.
    A parent path is a path that contains the given
    path at any relative depth.

    :type path: String
    :param path: The (base) path for the checking.
    :type parent_path: String
    :param parent_path: The path to be checked for
    parenting.
    :rtype: bool
    :return: The result of the checking for parenting.
    """

    # checks the relative path between the path and
    # the parent path
    _relative_path = relative_path(path, parent_path)

    # in case the relative path start with the
    # parent directory character
    if _relative_path.startswith(PARENT_DIRECTORY):
        # returns false (invalid)
        return False
    # otherwise it muse be a parent path
    else:
        # returns false (invalid)
        return True

def _relative_path_windows(path, start_path = CURRENT_DIRECTORY):
    """
    "Calculates" the relative path between the base path
    and the given "start" path.
    This version of the calculus is target at windows platforms.

    :type path: String
    :param path: The path to be used as "target".
    :type start_path: String
    :param start_path: The path to be used as starting point.
    :rtype: String
    :return: The relative path between both paths.
    """

    # in case the path is not defined (error)
    if not path:
        # raises a base value error
        raise ValueError("no path specified")

    # retrieves the various util values from the start
    # path and path
    start_is_unc, start_prefix, start_list = _abspath_split(start_path)
    path_is_unc, path_prefix, path_list = _abspath_split(path)

    # in case one of the paths is of type unc and the other
    # is not (error)
    if path_is_unc ^ start_is_unc:
        # raises a value error
        raise ValueError("cannot mix unc and non-unc paths %s and %s" % (path, start_path))

    if path_prefix.lower() != start_prefix.lower():
        if path_is_unc:
            # raises a value error
            raise ValueError("path is on unc root %s, start on unc root %s" % (path_prefix, start_prefix))
        else:
            # raises a value error
            raise ValueError("path is on drive %s, start on drive %s" % (path_prefix, start_prefix))

    # works out how much of the file path
    # is shared by start and path
    index = 0

    # iterates over both sub paths list to check for differences
    for start_sub_path, sub_path in zip(start_list, path_list):
        # in case the sub paths are different
        if start_sub_path.lower() != sub_path.lower():
            # breaks the loop
            break

        # increment the index
        index += 1

    # calculates the "relatives" list using the start list and path list
    # as reference
    relatives_list = [PARENT_DIRECTORY] * (len(start_list) - index) + path_list[index:]

    # in case the are not relatives list
    # (is the same directory)
    if not relatives_list:
        # returns the current directory
        return CURRENT_DIRECTORY

    # returns the result of joining all
    # the relatives list
    return os.path.join(*relatives_list)

def _relative_path_posix(path, start = CURRENT_DIRECTORY):
    """
    "Calculates" the relative path between the base path
    and the given "start" path.
    This version of the calculus is target at posix platforms.

    :type path: String
    :param path: The path to be used as "target".
    :type start_path: String
    :param start_path: The path to be used as starting point.
    :rtype: String
    :return: The relative path between both paths.
    """

    # in case the path is not defined (error)
    if not path:
        # raises a base value error
        raise ValueError("no path specified")

    # retrieves the sub paths list for both the start path and the path
    start_list = [value for value in os.path.abspath(start).split(SEPARATOR) if value]
    path_list = [value for value in os.path.abspath(path).split(SEPARATOR) if value]

    # works out how much of the file path is shared by start and path
    index = len(os.path.commonprefix([start_list, path_list]))

    # calculates the "relatives" list using the start list and path list
    # as reference
    relatives_list = [PARENT_DIRECTORY] * (len(start_list) - index) + path_list[index:]

    # in case the are not relatives list
    # (is the same directory)
    if not relatives_list:
        # returns the current directory
        return CURRENT_DIRECTORY

    # returns the result of joining all
    # the relatives list
    return os.path.join(*relatives_list)

def _abspath_split(path):
    """
    Method for util splitting of the path into
    is unc, prefix and path list.
    The values are returned in a tuple.

    :type path: String
    :param path: The path to be checked.
    :rtype: Tuple
    :return: A tuple containing the is unc, prefix and
    path list util values.
    """

    # normalizes the path and the retrieves
    # the absolute path and then splits it into
    # prefix and rest
    path = os.path.normpath(path)
    absolute_path = os.path.abspath(path)
    prefix, rest = os.path.splitunc(absolute_path) #@UndefinedVariable

    # converts the prefix to boolean for checking
    # if the path is unc based
    is_unc = bool(prefix)
    if not is_unc:
        # splits the absolute path for drive checking
        prefix, rest = os.path.splitdrive(absolute_path)

    # retrieves the list of various (sub) paths from the rest
    path_list = [value for value in rest.split(SEPARATOR) if value]

    # returns the tuple value
    return is_unc, prefix, path_list

# checks if the current platform is of type windows
if NT_PLATFORM_VALUE in sys.builtin_module_names or CE_PLATFORM_VALUE in sys.builtin_module_names:
    # sets the separator value and the relative
    # path method values (for windows)
    SEPARATOR = "\\"
    relative_path = _relative_path_windows
# otherwise it should be a posix environment
else:
    # sets the separator value and the relative
    # path method values (for posix)
    SEPARATOR = "/"
    relative_path = _relative_path_posix



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import math

from colony.base import legacy

FLOAT_PRECISION = 14
""" The amount of precision (in decimal places) that
is going to be used for the decimal internal usage """

class Decimal(float):
    """
    Fixed point rational number representation/manipulation
    structure aimed at replacing the internal python data
    structure with the same name.

    Provides a simple unified way of re-using the float data
    type to performed fixed point operations, using techniques
    around the special rounding method.

    The implementation of the data structure is not aimed at
    performance and because of that it should not be used for
    task that are considered performance intensive.
    """

    def __new__(self, value = 0.0):
        value = float(value)
        integer = abs(int(value // 1))
        count = 1 if integer == 0 else int(math.log10(integer)) + 1
        places = FLOAT_PRECISION - count
        self.places = places
        value = round(value, places)
        return super(Decimal, self).__new__(self, value)

    def __hash__(self):
        return float.__hash__(self)

    def __cmp__(self, other):
        other = self._normalize(other)
        if float.__gt__(self, other): return 1
        if float.__lt__(self, other): return -1
        return 0

    def __lt__(self, other):
        other = self._normalize(other)
        return float.__lt__(self, other)

    def __le__(self, other):
        other = self._normalize(other)
        return float.__le__(self, other)

    def __eq__(self, other):
        other = self._normalize(other)
        return float.__eq__(self, other)

    def __ne__(self, other):
        other = self._normalize(other)
        return float.__ne__(self, other)

    def __gt__(self, other):
        other = self._normalize(other)
        return float.__gt__(self, other)

    def __ge__(self, other):
        other = self._normalize(other)
        return float.__ge__(self, other)

    def __add__(self, other):
        result = float.__add__(self, other)
        return Decimal(result)

    def __radd__(self, other):
        result = float.__radd__(self, other)
        return Decimal(result)

    def __sub__(self, other):
        result = float.__sub__(self, other)
        return Decimal(result)

    def __rsub__(self, other):
        result = float.__rsub__(self, other)
        return Decimal(result)

    def __mul__(self, other):
        result = float.__mul__(self, other)
        return Decimal(result)

    def __rmul__(self, other):
        result = float.__rmul__(self, other)
        return Decimal(result)

    def __floordiv__(self, other):
        result = float.__floordiv__(self, other)
        return Decimal(result)

    def __rfloordiv__(self, other):
        result = float.__rfloordiv__(self, other)
        return Decimal(result)

    def __div__(self, other):
        result = float.__div__(self, other)
        return Decimal(result)

    def __rdiv__(self, other):
        result = float.__rdiv__(self, other)
        return Decimal(result)

    def __truediv__(self, other):
        result = float.__truediv__(self, other)
        return Decimal(result)

    def __rtruediv__(self, other):
        result = float.__rtruediv__(self, other)
        return Decimal(result)

    def __mod__(self, other):
        result = float.__mod__(self, other)
        return Decimal(result)

    def __rmod__(self, other):
        result = float.__rmod__(self, other)
        return Decimal(result)

    def __and__(self, other):
        other = self._normalize(other)
        return float.__and__(self, other)

    def __rand__(self, other):
        other = self._normalize(other)
        return float.__rand__(self, other)

    def __or__(self, other):
        other = self._normalize(other)
        return float.__or__(self, other)

    def __ror__(self, other):
        other = self._normalize(other)
        return float.__ror__(self, other)

    def __xor__(self, other):
        other = self._normalize(other)
        return float.__xor__(self, other)

    def __rxor__(self, other):
        other = self._normalize(other)
        return float.__rxor__(self, other)

    def __pos__(self):
        result = float.__pos__(self)
        return Decimal(result)

    def __neg__(self):
        result = float.__neg__(self)
        return Decimal(result)

    def __abs__(self):
        result = float.__abs__(self)
        return Decimal(result)

    def __invert__(self):
        result = float.__invert__(self)
        return Decimal(result)

    def __round__(self, n = 0):
        result = float.__round__(self, n)
        return Decimal(result)

    def __floor__(self):
        result = math.floor(float(self))
        return Decimal(result)

    def __ceil__(self):
        result = math.ceil(float(self))
        return Decimal(result)

    def __trunc__(self):
        result = float.__trunc__(self)
        return Decimal(result)

    def _normalize(self, value):
        if not type(value) == float: return value
        return round(value, self.places)

class JournaledList(list):
    """
    List structure that keeps track of the append and
    remove operation in a jounalized format.
    This structures is relevant for use cases where
    "diffs" around a base list must be kept.
    """

    _appends = []
    """ The list containing the various appends to the list (journal) """

    _removes = []
    """ The list containing the various removes from the list (journal) """

    def __init__(self, *args, **kwargs):
        """
        Constructor of the class, this constructor
        may be used together with a previously "simple" list
        to start the jounalized list with initial (non logged)
        values.
        """

        list.__init__(self, *args, **kwargs)

        self._appends = []
        self._removes = []

    def append(self, object):
        """
        Appends an object to the list, keeping the registry
        of the operation in the appends list.

        :type object: Object
        :param object: The object to be appended to the list.
        """

        # appends the object to the list
        list.append(self, object)

        # in case the object is present in the removes
        # list it must be removed (previous reverse operations
        # should reverted)
        if object in self._removes:
            # removes the object from the removes list
            # (reversion of operation)
            self._removes.remove(object)
        # otherwise the object must be added to the
        # appends list (normal behavior)
        else:
            # appends the object to the appends list
            # (for logging)
            self._appends.append(object)

    def remove(self, object):
        """
        Removes an object from the list, keeping the registry
        of the operation in the removes list.

        :type object: Object
        :param object: The object to be removed from the list.
        """

        # removes the object from the list, an exception
        # should be raises in case it fails
        list.remove(self, object)

        # in case the object is present in the appends
        # list it must be removed (previous reverse operations
        # should reverted)
        if object in self._appends:
            # removes the object from the appends list
            # (reversion of operation)
            self._appends.remove(object)
        # otherwise the object must be added to the
        # removes list (normal behavior)
        else:
            # appends the object to the removes list
            # (for logging)
            self._removes.append(object)

    def clear_jounal(self):
        """
        Clears the journal, reseting it to the original
        state (internal structures state).
        This method should be called whenever a new jounalized
        unit is required for a new phase
        """

        del self._appends[:]
        del self._removes[:]

    def get_appends(self):
        """
        Retrieves the list of the current (valid) append operations
        from the journaled list.

        :rtype: List
        :return: The list of the current (valid) append operations
        from the journaled list.
        """

        return self._appends

    def get_removes(self):
        """
        Retrieves the list of the current (valid) remove operations
        from the journaled list.

        :rtype: List
        :return: The list of the current (valid) remove operations
        from the journaled list.
        """

        return self._removes

    def _append(self, object):
        """
        Appends an object to the list, avoiding the keeping
        of a log on that operation (simple operation).

        :type object: Object
        :param object: The object to be appended to the list.
        """

        list.append(self, object)

    def _remove(self, object):
        """
        Removes an object from the list, avoiding the keeping
        of a log on that operation (simple operation).

        :type object: Object
        :param object: The object to be removed from the list.
        """

        list.remove(self, object)

class OrderedMap(object):
    """
    Structure that allow the usage of a map
    like syntax to create ordered elements.

    The ordered map uses a composing strategy to
    achieve the extra behavior for order in map.
    """

    tuples_list = None
    """ The list of tuples """

    _map = None
    """ The map to be used internally for virtual access """

    _keys = None
    """ The internal keys list for ordered keys retrieval """

    def __init__(self, ordered_keys = False, map = None):
        """
        Constructor of the class.

        In case a map is provided the map is first sorted according
        to the default sorting operation in the map keys and then
        all of their values are entered.

        :type ordered_keys: bool
        :param ordered_keys: If the keys should also be provided
        in an ordered fashion (expensive remove operation).
        :type map: Dictionary
        :param map: The map to be used to populate the ordered map
        initially using the default key order.
        """

        self.tuples_list = []
        self._map = {}

        if ordered_keys: self._keys = []
        if map: self._load_map(map)

    def __len__(self):
        return self._map.__len__()

    def __getitem__(self, key):
        return self._map[key]

    def __setitem__(self, key, value):
        self.__add_item(key, value)

    def __delitem__(self, key):
        self.__remove_item(key)

    def __iter__(self):
        return OrderedMapIterator(self)

    def __contains__(self, item):
        return self._map.__contains__(item)

    def get(self, key, default_value):
        return self._map.get(key, default_value)

    def values(self):
        return self._map.values()

    def items(self):
        return self.tuples_list

    def extend(self, map):
        # iterates over all the map items
        for key, value in legacy.iteritems(map):
            # sets the item in the structure
            self.__setitem__(key, value)

    def keys(self):
        # in case the keys (list) is not
        # defined (no ordered keys used)
        if self._keys == None:
            keys = self._map.keys()
        # otherwise the ordered keys list
        # is available and must be used
        else:
            keys = self._keys

        # returns the valid keys
        return keys

    def itervalues(self):
        return self._map.itervalues()

    def iteritems(self):
        return self.items()

    def iterkeys(self):
        return self.keys()

    def _load_map(self, map):
        """
        Loads the provided map into the current ordered
        map so that the keys are sorted in the default
        manner and then the associated items are inserted
        using the order of the keys.

        :type map: Dictionary
        :param map: The map that is going to be used to
        populate the current ordered map using the default
        order of the keys.
        """

        keys = map.keys()
        keys.sort()
        for key in keys:
            value = map[key]
            self.__add_item(key, value)

    def __add_item(self, key, value):
        """
        Adds an item with the given key to the
        internal structures.

        :type key: String
        :param key: The key of the element to be added
        to the internal structures.
        :type value: Object
        :param value: The value of the element to be added
        to the internal structures.
        """

        # in case the key (item) exists
        # in the map
        if key in self._map:
            # removes the item from the internal
            # structures (using the key)
            self.__remove_item(key)

        # adds the tuple to the tuples list
        self.tuples_list.append((key, value))

        # sets the value in the map
        self._map[key] = value

        # adds the key to the keys list (only in case
        # the keys list is available and set and the key
        # is not present in the keys list)
        if not self._keys == None and not key in self._keys: self._keys.append(key)

    def __remove_item(self, key):
        """
        Removes the item with the given key from the
        internal structures.

        :type key: String
        :param key: The key of the element to be removed
        from the internal structures.
        """

        # retrieves the value for the key
        value = self._map[key]

        # removes the tuple from the tuples list
        self.tuples_list.remove((key, value))

        # removes the value from the map
        del self._map[key]

        # removes the key from the keys list (only in case
        # the keys list is available and set)
        if not self._keys == None: self._keys.remove(key)

class OrderedMapIterator(object):
    """
    The iterator for the ordered map.
    """

    ordered_map = None
    """ The ordered map to be used """

    current_index = None
    """ The current index value """

    def __init__(self, ordered_map):
        """
        Constructor of the class.

        :type ordered_map: OrderedMap
        :param ordered_map: The ordered map to be used by the iterator.
        """

        self.ordered_map = ordered_map

        self.current_index = 0

    def __next__(self):
        """
        Retrieves the next ordered map key.

        This is the magic method responsible for the retrieval of
        the next value (should forward the request).

        :rtype: String
        :return: The next key in the ordered map.
        """

        return self.next()

    def next(self):
        """
        Retrieves the next ordered map key.

        :rtype: String
        :return: The next key in the ordered map.
        """

        # retrieves the ordered map length
        ordered_map_length = len(self.ordered_map)

        # in case there is an overflow (current index is over the length)
        if not ordered_map_length > self.current_index:
            # breaks the iteration
            raise StopIteration()

        # retrieves the key and value for the current index
        key, _value = self.ordered_map.tuples_list[self.current_index]

        # increments the current index
        self.current_index += 1

        # returns the current key value
        return key

class MultipleValueMap(object):
    """
    Map that holds multiple values for
    each key, and considers
    the first value to the key's value.
    """

    _map = None
    """ The map to be used internally for virtual access """

    def __init__(self):
        """
        Constructor of the class.
        """

        # initializes the map
        self._map = {}

    def __len__(self):
        return self._map.__len__()

    def __getitem__(self, key):
        # returns the value
        values = self._map.get(key)

        # returns in case no
        # value was found
        if not values:
            return

        # retrieves the first value
        value = values[0]

        # returns the value
        return value

    def __setitem__(self, key, value):
        # retrieves the values
        values = self._map.get(key, [])

        # adds the value to the list
        values.append(value)

        # sets the values in the map
        self._map[key] = values

    def __delitem__(self, key):
        del self._map[key]

    def __iter__(self):
        return self._map.__iter__()

    def __contains__(self, item):
        return self._map.__contains__(item)

    def get(self, key, default_value):
        return self._map.get(key, default_value)

    def values(self):
        return self._map.values()

    def items(self):
        return self.tuples_list

    def keys(self):
        return self._map.keys()

    def itervalues(self):
        return self._map.itervalues()

    def iteritems(self):
        return self.items()

    def iterkeys(self):
        return self._map.keys()

    def unset(self, key, value):
        """
        Unsets the value from the specified key.

        :type key: Object
        :param key: The key where to unset the value.
        :type value: Object
        :param value: The value to unset.
        """

        # retrieves the values
        values = self._map[key]

        # removes the value
        values.remove(value)

class FormatTuple(object):
    """
    Tuple based structure that may be used to represent
    a string to be formated with a series of values.
    This structure provides a portable way of passing
    the base format string and a series of arguments.
    """

    format_string = None
    """ The (base) format string to be used in the
    process of formatting the final string """

    arguments = ()
    """ The tuple containing the various arguments to
    be used during the formatting of the string """

    def __init__(self, format_string, *args):
        """
        Constructor of the class.

        :type format_string: String
        :param format_string: The format string to be
        used during the process of formatting.
        """

        self.format_string = format_string
        self.arguments = args

    @staticmethod
    def build(format_string, *args):
        """
        Builds (constructs) a new format tuple from the provided
        format string and (optional) series of arguments for the
        formatting.

        :type format_string: String
        :param format_string: The string to be used for formatting of
        the string result.
        :rtype: FormatTuple
        :return: The resulting format tuple object from the
        building "generated" from the provided arguments.
        """

        return FormatTuple(format_string, *args)

    def __hash__(self):
        return hash(self.format_string)

    def __str__(self):
        return self.format()

    def __eq__(self, other):
        if other == self.format_string: return True
        return False

    def __ne__(self, value):
        return not self.__eq__(value)

    def __add__(self, other):
        return self.format() + other

    def __radd__(self, other):
        return other + self.format()

    def __replace__(self, value):
        self.set_format_string(value)

    def json_v(self, format = False):
        if format: return self.format()
        else: return (self.format_string, self.arguments)

    def format(self, format_string = None):
        """
        Formats the current format tuple using the currently loaded
        attributes and the provided format string.

        In case no format string is provided the default format
        string is used instead.

        :type format_string: String
        :param format_string: The string to be used for formatting of
        the final string result.
        :rtype: String
        :return: The "final" resulting string after the format of it.
        """

        # retrieves the "final" format string, using the
        # provided format string or the "original" and
        # default format string
        format_string = format_string or self.format_string

        # uses the format string to format the string using
        # the currently available arguments, then returns
        # the resulting string value
        return format_string % self.arguments

    def get_format_string(self):
        """
        Retrieves the currently associated string for formatting
        in the current format tuple.

        :rtype: String
        :return: The currently available format string.
        """

        return self.format_string

    def set_format_string(self, format_string):
        """
        Sets the currently associated string for formatting
        in the current format tuple.

        :type: String
        :param: The string to be set as the currently
        available format string.
        """

        self.format_string = format_string

class FileReference(object):
    """
    Simple structure for the description of a file in the
    current file system.

    The structure contains a series of utility methods that
    allow fast and simple operation on the described file.
    """

    path = None
    """ The file to the file that is being described by the
    the current file reference object """

    encoding = None
    """ The encoding used for the file in case the referred
    file is text based, useful for simple reading operations """

    def __init__(self, path, encoding = None):
        """
        Constructor of the class.

        :type path: String
        :param path: The path to the file that is being described
        by the current file reference object to be created.
        :type encoding: String
        :param encoding: String describing the encoding used by the
        referenced file (only for text based).
        """

        self.path = path
        self.encoding = encoding

    def read_all(self, mode = "rb"):
        """
        Reads the complete set of the contents from the file that
        is currently being described.

        :type mode: String
        :param mode: The file open mode to be used for the reading
        (may control binary/text modes).
        """

        file = open(self.path, mode)
        try: data = file.read()
        finally: file.close()
        if self.encoding: data = data.decode(self.encoding)
        return data

def is_dictionary(object):
    """
    Validates if the given object is a valid
    dictionary object.

    :type object: Object
    :param object: The object to be validated.
    :rtype: bool
    :return: If the given object is a valid
    dictionary object.
    """

    # retrieves the object type
    object_type = type(object)

    # in case the object type is dictionary
    if object_type == dict: return True

    # in case the object type id ordered map
    # (this is the custom dictionary class)
    if object_type == OrderedMap: return True

    # returns false
    return False



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import logging
import itertools
import collections

from . import legacy

try:
    import zmq
    boradcast = True
except ImportError:
    boradcast = False

MAX_LENGTH = 10000
""" The maximum amount of messages that are kept in
memory until they are discarded, avoid a very large
number for this value or else a large amount of memory
may be used for logging purposes """

LEVELS = (
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL"
)
""" The sequence of levels from the least sever to the
most sever this sequence may be used to find all the
levels that are considered more sever that a level """

LEVEL_ALIAS = dict(
    DEBU = "DEBUG",
    WARN = "WARNING",
    INF = "INFO",
    ERR = "ERROR",
    CRIT = "CRITICAL"
)
""" Map defining a series of alias that may be used latter
for proper debug level resolution, standard compliant """

class BroadcastHandler(logging.Handler):
    """
    Logging handler that broadcasts the messages using the
    ZeroMQ technology. It's useful for distributed logging
    situations where one wants to gather logging from a
    colony instance spread over a network space.

    :see: https://zeromq.org
    """

    socket = None
    """ The current ZeroMQ socket in used, this is
    the object to be used to broadcast the messages """

    def __init__(self, host = None, port = None):
        logging.Handler.__init__(self)

        if not boradcast: return

        context = zmq.Context()
        self.socket = context.socket(zmq.PUB)
        self.socket.bind("tcp://127.0.0.1:5600")

    def flush(self):
        """
        Flushes the stream, this should be able to send
        the pending contents in the stream through it.
        """

        pass

    def emit(self, record):
        """
        Emit a record, to the output stream the record will
        be formated and then sent to the socket for broadcast.

        :type record: Record
        :param record: The log record to be sent through the
        output socket stream.
        """

        # in case the broadcast flag is unset returns
        # immediately can't be used
        if not boradcast: return

        try:
            # formats the record, retrieving the resulting
            # message string
            message = self.format(record)

            # in case the type of the message is unicode encodes it using the
            # default encoding for the transmission, then sends the message
            # through the socket
            if type(message) == legacy.UNICODE: message = message.encode("utf-8")
            self.socket.send(b"colony %s" % message)

            # flushes the current stream
            self.flush()
        except Exception:
            # handles the error in the proper manner
            self.handleError(record)

class MemoryHandler(logging.Handler):
    """
    Logging handler that is used to store information in
    memory so that anyone else may consult it latter as
    long as the execution session is the same.
    """

    def __init__(self, level = logging.NOTSET, max_length = MAX_LENGTH):
        logging.Handler.__init__(self, level = level)
        self.max_length = max_length
        self.messages = collections.deque()
        self.messages_l = dict()

    def get_messages_l(self, level):
        # in case the level is not found in the list of levels
        # it's not considered valid and so an empty list is returned
        try: index = LEVELS.index(level)
        except Exception: return collections.deque()

        # retrieves the complete set of levels that are considered
        # equal or more severe than the requested one
        levels = LEVELS[:index + 1]

        # creates the list that will hold the various message
        # lists associated with the current severity level
        messages_l = collections.deque()

        # iterates over the complete set of levels considered
        # equal or less sever to add the respective messages
        # list to the list of message lists
        for level in levels:
            _messages_l = self.messages_l.get(level, None)
            if _messages_l == None: _messages_l = collections.deque()
            self.messages_l[level] = _messages_l
            messages_l.append(_messages_l)

        # returns the complete set of messages lists that
        # have a level equal or less severe that the one
        # that has been requested by argument
        return messages_l

    def emit(self, record):
        # formats the current record according to the defined
        # logging rules so that we can used the resulting message
        # for any logging purposes
        message = self.format(record)

        # retrieves the level (as a string) associated with
        # the current record to emit and uses it to retrieve
        # the associated messages list
        level = record.levelname
        messages_l = self.get_messages_l(level)

        # inserts the message into the messages queue and in
        # case the current length of the message queue overflows
        # the one defined as maximum must pop message from queue
        self.messages.appendleft(message)
        messages_s = len(self.messages)
        if messages_s > self.max_length: self.messages.pop()

        # iterates over all the messages list included in the retrieve
        # messages list to add the logging message to each of them
        for _messages_l in messages_l:
            # inserts the message into the proper level specific queue
            # and in case it overflows runs the same pop operation as
            # specified also for the more general queue
            _messages_l.appendleft(message)
            messages_s = len(_messages_l)
            if messages_s > self.max_length: _messages_l.pop()

    def clear(self):
        self.messages = collections.deque()
        self.messages_l = dict()

    def get_latest(self, count = None, level = None):
        count = count or 100
        is_level = level and not legacy.is_string(level)
        if is_level: level = logging.getLevelName(level)
        level = level.upper() if level else level
        level = LEVEL_ALIAS.get(level, level)
        messages = self.messages_l.get(level, []) if level else self.messages
        slice = itertools.islice(messages, 0, count)
        return list(slice)

    def flush_to_file(
        self,
        path,
        count = None,
        level = None,
        reverse = True,
        clear = True
    ):
        messages = self.get_latest(level = level, count = count or 65536)
        if not messages: return
        if reverse: messages.reverse()
        is_path = isinstance(path, legacy.STRINGS)
        file = open(path, "wb") if is_path else path
        try:
            for message in messages:
                message = legacy.bytes(message, "utf-8", force = True)
                file.write(message + b"\n")
        finally:
            if is_path: file.close()
        if clear: self.clear()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import xml.dom.minidom

from colony.base import legacy

def xml_to_dict(data):
    """
    Converts the provided linear XML string value into
    a dictionary that represents the same hierarchy.

    The conversion strategy is only applicable to simple
    hierarchical XML structures.

    :type data: String/Node
    :param data: The XML string (or XML node) to be used in
    the (to dictionary) conversion.
    :rtype: Dictionary
    :return: The dictionary representation of the XML data.
    """

    if isinstance(data, xml.dom.Node): node = data
    else: node = xml.dom.minidom.parseString(data)
    return _node_to_dict(node)

def dict_to_xml(contents, encoding = "utf-8"):
    """
    Converts the provided dictionary structure into an
    XML string string in a recursive fashion.

    :type contents: Dictionary
    :param contents: The dictionary that is going to be
    converted into a linear XML string.
    :type encoding: String
    :param encoding: The string encoding to be used if
    necessary to decode byte base string elements.
    :rtype: String
    :return: The final XML string value representing the
    given dictionary in a linear fashion.
    """

    buffer = []
    for key in sorted(legacy.keys(contents)):
        value = contents[key]
        if isinstance(value, dict):
            value = dict_to_xml(value, encoding = encoding)
        elif legacy.is_bytes(value):
            value = value.decode(encoding)
        elif value == None:
            value = legacy.u("")
        buffer.append(legacy.u("<%s>%s</%s>") % (key, value, key))
    return legacy.u("").join(buffer)

def _node_to_dict(node):
    contents = None
    for _node in node.childNodes:
        if _node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if not contents: contents = dict()
            contents[_node.nodeName] = _node_to_dict(_node)
        if _node.nodeType == xml.dom.Node.TEXT_NODE and _node.data.strip():
            contents = _node.data.strip()
    return contents



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys
import copy
import calendar
import datetime

from colony.base import legacy

def map_clean(map):
    """
    Cleans the map from all of its entries.
    The clean process is "slow" as it iterates over
    all the map keys to remove its values.

    :type map: Dictionary
    :param map: The map to be cleaned.
    """

    # retrieves the map keys
    map_keys = legacy.keys(map)

    # iterates over all the map keys deleting
    # the complete set of item present in it
    for map_key in map_keys: del map[map_key]

def map_get(map, keys = []):
    """
    Recursively retrieves values from the provided map using
    the sequence of keys provided.

    This method may suffer from stack overflow (recursion) so
    a reasonable list of keys should be provided to avoid that.

    :type map: Dictionary
    :param map: The map (containing maps) from which the value
    is going to be retrieved using a recursive strategy.
    :type keys: List
    :param keys: The list of keys to be used in the recursive
    retrieval of values from the map.
    :rtype: Object
    :return: The final value retrieved through recursion
    """

    if not keys: return map
    return map_get(map[keys[0]], keys[1:])

def map_copy(source_map, destiny_map):
    """
    Copies the contents of the source map to the destiny map.

    Note that in case the value already exists in the destiny
    map the copy step will be ignored.

    :type source_map: Dictionary
    :param source_map: The source map of the copy.
    :type destiny_map: Dictionary
    :param destiny_map: The destiny map of the copy.
    """

    # iterates over all the source map keys in order to copy
    # the complete set of values from the source to destiny map
    for source_map_key in source_map:
        # retrieves the source map value, that may be set
        # in the destiny map in case no duplicate key exists
        source_map_value = source_map[source_map_key]

        # in case the key is not present in the destiny map
        # must add it into the destiny map (ensures existence)
        if not source_map_key in destiny_map or\
            destiny_map[source_map_key] == None or\
            destiny_map[source_map_key] == "none":
            destiny_map[source_map_key] = source_map_value

def map_copy_deep(source_map, destiny_map):
    """
    Copies the contents of the source map to the destiny map.
    This mode provides a deep copy, using a recursive approach.

    :type source_map: Dictionary
    :param source_map: The source map of the copy.
    :type destiny_map: Dictionary
    :param destiny_map: The destiny map of the copy.
    """

    # copies the current source map to the destiny map
    map_copy(source_map, destiny_map)

    # iterates over all the source map items
    for source_key, source_value in legacy.iteritems(source_map):
        # retrieves the source value type
        source_value_type = type(source_value)

        # in case the source value type is not a dictionary
        # continues the loop, nothing to be done in the
        # current iteration
        if not source_value_type == dict: continue

        # creates the destiny value map and sets the
        # destiny value in the destiny map
        destiny_value = {}
        destiny_map[source_key] = destiny_value

        # copies the source value (map) to the destiny value
        map_copy_deep(source_value, destiny_value)

def map_duplicate(item):
    """
    Duplicates the provided item (map) creating a new
    structure with duplicated references both for sequences
    (list and tuples) and for maps.

    This function is useful in order to avoid reference
    overlapping in data structures.

    :type item: Object
    :param item: The item to be used as reference for duplication
    this should be a map at the initial call of the function.
    :rtype: item: Object
    :return: The duplicated data structure with all the references
    replicated in the sequences and maps.
    """

    # retrieves the type for the current
    # item in order to percolate it appropriately
    _type = type(item)

    # in case the current item is a sequence
    # must "copy" all of its elements
    if _type in (list, tuple):
        return [map_duplicate(value) for value in item]

    # in case the current item is a map
    # must create a new map with the result
    # of the copy of all the elements
    elif _type == dict:
        # creates the item map to be populated
        # with the copied values
        _item = {}

        # iterates over all the items to copy
        # them and sets them in the new items map
        for key, value in legacy.iteritems(item):
            _item[key] = map_duplicate(value)
        return _item

    # otherwise must be a "single" item and the
    # (reduce) operation must be performed on it
    else:
        return item

def map_remove(removal_map, destiny_map):
    """
    Removes all the values with keys present in the
    removal map from the destiny map.

    :type removal_map: Dictionary
    :param removal_map: The map to be used in reference
    with the key values.
    :type destiny_map: Dictionary
    :param destiny_map: The "destiny" map to have the values removed.
    """

    # iterates over all the keys in
    # the removal map
    for key in removal_map:
        # in case the key does not exists in
        # destiny map, continues the loop
        if not key in destiny_map: continue

        # removes the key item from the destiny map
        del destiny_map[key]

def map_extend(
    base_map,
    extension_map,
    override = True,
    recursive = False,
    copy_base_map = True
):
    """
    Extends the given map with the extension map,
    retrieving a map resulting of the merge of both maps.

    In case the override flag is set the values are overridden
    in case they already exist in the base map.

    An optional recursive flag allows the extension to be
    made recursive in case the value is a map.

    The base map may be changed or left untouched based on
    the copy base map flag.

    :type base_map: Dictionary
    :param base_map: The map to be used as base for the merge,
    should contain more values at the end of the merge.
    :type extension_map: Dictionary
    :param extension_map: The map to be used to extend the base
    one, this map may be recursively percolated if requested.
    :type override: bool
    :param override: If a value should be overridden in
    case it already exists in the base map.
    :type recursive: bool
    :param recursive: If a value should be extended in
    case it already exists in the base map (recursive
    extension).
    :type copy_base_map: bool
    :param copy_base_map: If the base map should be copied before
    being extended in order to avoid loss of data.
    :rtype: Dictionary
    :return: The map that result of the merge of both maps.
    """

    # copies the base map to create the initial result map (optional)
    result_map = copy.copy(base_map) if copy_base_map else base_map

    # iterates over all the keys and values
    # in the extension map
    for key, value in legacy.iteritems(extension_map):
        # in case the override flag is not set and
        # the key already exists in the result map
        # must skip the current iteration loop
        if not override and key in result_map:
            continue

        # retrieves the data type for the current value in
        # iteration so that it may be used to determine the
        # need for the recursive percolated copy/merge
        value_type = type(value)

        # in case the value is a map and the recursive flag
        # is set, must try to extend the current item with
        # the new one, as expected
        if recursive and value_type == dict:
            #  tries to retrieve the result map value, so
            # that it's possible to extend the current value
            # with the previously existing one, a default map
            # is created in case it does not exists so that
            # it's possible to used as the base for population
            result_map_value = result_map.get(key, {})

            # extends the map that is currently in the result
            # map with the new value (recursive step), note
            # that a new map may have been created to be used
            # as the placeholder of new map values, this avoids
            # collision between sub-maps (would cause problems)
            value = map_extend(
                result_map_value,
                value,
                override = override,
                recursive = recursive,
                copy_base_map = copy_base_map
            )

        # sets the (extension) value in the result map
        result_map[key] = value

    # returns the result map
    return result_map

def map_flatten(map):
    """
    Flattens the provided map, meaning that in case there's a
    value in the map associated with another map the name in
    that map will be linearized into the top level map.

    :type map: Dictionary
    :param map: The map that is going to be linearized/flatten
    and for each relations will be set in the top level.
    :rtype: Dictionary
    :return: The resulting flatten map, note that the original
    map is not changed.
    """

    # creates a new dictionary/map to hold the resulting map,
    # then retrieves the various linear key and value pairs
    # for the complete set of relations setting the values
    # in the resulting map and returns the map to caller method
    result = dict()
    pairs = _map_flatten_pairs(map)
    for key, value in pairs: result[key] = value
    return result

def map_check_parameters(map, parameters_list, exception = Exception):
    """
    Checks if the parameters in the parameters list are defined
    in the given map.
    In case a check fails an exception is raised.

    :type map: Dictionary
    :param map: The dictionary to be checked.
    :type parameters_list: List
    :param parameters_list: The list of parameters to be checked.
    :type exception: Exception
    :param exception: The exception to be raised in case check fails.
    """

    # iterates over all the parameters in the parameters list
    # to try to verify for their existence in the map
    for parameter in parameters_list:
        # in case the parameter is in the map must continue
        # the current loop, no exception raised this time
        if parameter in map: continue

        # raises the provided exception as the parameter does
        # not exist in the provided map
        raise exception(parameter)

def map_get_value_cast(map, key, cast_type = str, default_value = None):
    """
    Retrieves the value of the map for the given key.
    The value is casted to the given type.
    In case something wrong (exception raised) occurs
    the default value is returned.

    :type map: Dictionary
    :param map: The dictionary to be used.
    :type key: String
    :param key: The key to the value to be retrieved.
    :type cast_type: Type
    :param cast_type: The type to be used to cast the retrieved
    value (this should be a valid type, with constructor).
    :type default_value: Object
    :param default_value: The default value to be used
    when something wrong (exception raised) occurs.
    :rtype: Object
    :return: The retrieved value casted to the defined type.
    """

    try:
        # retrieves the value for the map
        value = map[key]

        # retrieves the value type
        value_type = type(value)

        # in case the type of the value
        # is the "target" cast type
        if value_type == cast_type:
            # sets the value casted as the
            # "original" value
            value_casted = value
        # otherwise
        else:
            # casts the value for the cast type
            value_casted = cast_type(value)

        # returns the value casted
        return value_casted
    except Exception:
        # returns the default value
        return default_value

def map_get_values(map, key):
    """
    Retrieves the value of the map for the given key.
    The value is validated for type and in case it's not
    a list a list is created with the value.
    This way the method return value is always a list
    independently from the type of the original value.

    :type map: Dictionary
    :param map: The dictionary to be used.
    :type key: String
    :param key: The key to the value(s) to be retrieved.
    :rtype: List
    :return: The values(s) for the key.
    """

    # retrieves the values from the map
    values = map.get(key, [])

    # retrieves the values type
    values_type = type(values)

    # in case the values element is not a list
    # the values element is converted into a
    # a list with the original values in it,
    # ensuring that the returned value is always
    # a valid sequence value, ready to be iterated
    if not values_type == list: values = [values]

    # returns the values
    return values

def map_output(map, output_method = sys.stdout.write, indentation = ""):
    """
    Outputs (pretty print) the given map, using the
    defined output method.
    The map is printed with the given indentation as a start
    point from the line.

    :type map: Dictionary
    :param map: The map to be outputted (pretty print).
    :type output_method: Function
    :param output_method: The output function (method) to be
    used for outputting the value.
    :type indentation: String
    :param indentation: The indentation level to be used.
    """

    # iterates over all the keys in the map in order to render
    # their value to the provided output method/function
    for key in map:
        # retrieves the map value
        map_value = map[key]

        # outputs the map value, meaning that a key to value
        # association will be "printed" to the output method
        if type(map_value) == dict:
            # defines the key string with the proper indentation
            # value and the key value and outputs it properly
            key_string = indentation + legacy.UNICODE(key) + ":"
            output_method(key_string)

            # defines the "new" output map indentation and runs the
            # recursive operation so that the new map is printed
            output_map_identation = indentation + "  "
            map_output(map_value, output_method, output_map_identation)

        # otherwise it must be a "simple" value and so the a "simple"
        # key and value string is going to be printed to the method
        else:
            # creates a string representation of the map value and
            # then "runs" it to the output method
            map_value_string = indentation + legacy.UNICODE(key) +\
                " = " + legacy.UNICODE(map_value)
            output_method(map_value_string)

def map_normalize(item, operation = None):
    """
    Normalizes the provided map/item, applying the reduce
    operation to each of the items.

    In case no operation is provided, the default reduce
    map operation is used.

    This operation may be used to convert "complex" type
    based maps into simplified type based maps.

    :type item: Object
    :param item: The map/item to be normalized.
    :type operation: Method
    :param operation: The operation used for normalization
    (reduce operation).
    :rtype: Object
    :return: The normalized map, resulting from the normalization
    of each of its items.
    """

    # sets the (reduce) operation, defaulting
    # to reduce map in case none is defined
    operation = operation or _map_reduce

    # retrieves the type for the current
    # item in order to percolate it appropriately
    _type = type(item)

    # in case the current item is a sequence
    # must normalize all of its elements
    if _type in (list, tuple):
        return [map_normalize(value, operation) for value in item]

    # in case the current item is a map
    # must create a new map with the result
    # of the normalization of all the elements
    elif _type == dict:
        # creates the item map to be populated
        # with the normalized values
        _item = {}

        # iterates over all the items to normalize
        # them and sets them in the new items map
        for key, value in legacy.iteritems(item):
            _item[key] = map_normalize(value, operation)
        return _item

    # otherwise must be a "single" item and the
    # (reduce) operation must be performed on it
    else:
        return operation(item)

def _map_flatten_pairs(map):
    """
    Retrieves the complete set of linear key to value pairs
    of the attributes of the provided map.

    Any dictionary based value of the provided map will be
    linearized using an extra recursion step.

    :type map: Dictionary
    :param map: The map that is the basis of the recursion
    step for the retrieval of the key value pairs.
    :rtype: Generator
    :return: A generator that yields the various key to value
    linear relations, using a recursive approach.
    """

    # iterates over the complete set of key value pairs of the
    # provided map to verify if the pair is map based and if
    # that's the case run a new depth of recursion with a new
    # composite name, so that all relations are linearized
    for key, value in legacy.iteritems(map):
        is_class = hasattr(value, "__class__")
        is_map = is_class and issubclass(value.__class__, dict)
        if not is_map: yield key, value; continue
        pairs = _map_flatten_pairs(value)
        for _key, _value in pairs: yield key + "." + _key, _value

def _map_reduce(value):
    """
    Reduces the provided value, converting it into the appropriate
    value considered to be "raw".

    :type value: Object
    :param value: The value to be reduced.
    :rtype: Object
    :return: The reduced value affected by the transformed function.
    """

    # retrieves the type for the item
    _type = type(value)

    # in case the current value is of type datetime then
    # it must be converted to the corresponding timestamp
    if _type == datetime.datetime:
        timetuple = value.utctimetuple()
        return calendar.timegm(timetuple)

    # returns the original value (default fallback)
    return value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2022 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2022 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import colony

class DecimalTest(colony.ColonyTestCase):
    """
    Class that tests the decimal (fixed point) data structure
    so that it ensured to be conformant with required operations.
    """

    def test_basic(self):
        """
        Runs a series of basic sanity tests for the decimal
        data structure.
        """

        result = +colony.Decimal(12.2)
        self.assertEqual(result, +12.2)
        self.assertEqual(type(result), colony.Decimal)

        result = -colony.Decimal(12.2)
        self.assertEqual(result, -12.2)
        self.assertEqual(type(result), colony.Decimal)

        result = abs(colony.Decimal(-12.2))
        self.assertEqual(result, 12.2)
        self.assertEqual(type(result), colony.Decimal)

        result = round(colony.Decimal(88.151), 2)
        self.assertEqual(result, 88.15)
        self.assertEqual(type(result), colony.Decimal)

    def test_arithmetic(self):
        """
        Tests a series of arithmetic operations around the
        decimal data structure (critical operations).
        """

        result = colony.Decimal(12.2) * 34.23
        self.assertEqual(result, 417.606)

        result = colony.Decimal(12687.23) * 34.132
        self.assertEqual(result, 433040.53436)

        result = colony.Decimal(532687.23) * 4534.23
        self.assertEqual(result, 2415326418.8829)

        result = 12.2 * 34.23
        self.assertNotEqual(result, 417.606)

        result = colony.Decimal(88.151) - 88.15
        self.assertEqual(result, 0.001)

        result = 88.151 - 88.15
        self.assertNotEqual(result, 0.001)

        result = colony.Decimal(888888888.151) - 88.55
        self.assertEqual(result, 888888799.601)

        result = 888888888.151 - 88.55
        self.assertNotEqual(result, 888888799.601)

        result = colony.Decimal(88.151) - 823.35
        self.assertEqual(result, -735.199)

        result = 88.151 - 823.35
        self.assertNotEqual(result, -735.199)

        result = colony.Decimal(88.151) - 823.35
        self.assertEqual(type(result), colony.Decimal)

        result = 88.151 - 823.35
        self.assertEqual(type(result), float)

        result = colony.Decimal(1.55)
        self.assertEqual(result, 1.55)

        result = colony.Decimal(9999999999.155555555555555555)
        self.assertEqual(result, 9999999999.15555555555555555)

        result = colony.Decimal(88.151) / 1
        self.assertEqual(result, 88.151)
        self.assertEqual(type(result), colony.Decimal)

        result = sum([colony.Decimal(88.151)])
        self.assertEqual(result, 88.151)
        self.assertEqual(type(result), colony.Decimal)

        result = 88 - colony.Decimal(88.15)
        self.assertEqual(result, -0.15)
        self.assertEqual(type(result), colony.Decimal)

        result = 88.151 - colony.Decimal(88.15)
        self.assertEqual(result, 0.001)
        self.assertEqual(type(result), colony.Decimal)

        result = colony.Decimal(3348.5) * colony.Decimal(0.23)
        self.assertEqual(result, 770.155)
        self.assertEqual(type(result), colony.Decimal)

        result = colony.rounds(result, 2)
        self.assertEqual(result, 770.16)
        self.assertEqual(type(result), colony.Decimal)

    def test_boolean(self):
        """
        Runs a series of tests on the boolean based operators
        defined for the decimal data structure.
        """

        result = colony.Decimal(12.2) and 12.2
        self.assertEqual(result, 12.2)
        self.assertEqual(type(result), float)

        result = 12.2 and colony.Decimal(12.2)
        self.assertEqual(result, colony.Decimal(12.2))
        self.assertEqual(type(result), colony.Decimal)

        result = colony.Decimal(12.2) or 12
        self.assertEqual(result, colony.Decimal(12.2))
        self.assertEqual(type(result), colony.Decimal)

    def test_operations(self):
        """
        Runs a series of general operations testing for the
        decimal data structure.
        """

        key = colony.Decimal(88.151) - 88.15
        self.assertEqual(type(key), colony.Decimal)

        map = dict()
        map[key] = "value"
        result = map[key]
        self.assertEqual(result, "value")

    def test_coercing(self):
        """
        Runs a series of coercing tests for the decimal data
        type conversion to other types and vice-versa
        """

        result = int(colony.Decimal(12.2))
        self.assertEqual(result, 12)
        self.assertEqual(type(result), int)

        result = int(colony.Decimal(12.99))
        self.assertEqual(result, 12)
        self.assertEqual(type(result), int)

        result = float(colony.Decimal(12.99))
        self.assertEqual(result, 12.99)
        self.assertEqual(type(result), float)

        result = str(colony.Decimal(12.99))
        self.assertEqual(result, "12.99")
        self.assertEqual(type(result), str)

        result = colony.Decimal(int(12))
        self.assertEqual(result, 12.0)
        self.assertEqual(type(result), colony.Decimal)

        result = colony.Decimal(float(12.99))
        self.assertEqual(result, 12.99)
        self.assertEqual(type(result), colony.Decimal)

        result = colony.Decimal("12.99")
        self.assertEqual(result, 12.99)
        self.assertEqual(type(result), colony.Decimal)

class JournaledListTest(colony.ColonyTestCase):
    """
    Class that tests the journaled list structure.
    """

    def test_append(self):
        """
        Tests the append method of the journaled list.
        """

        # creates a journaled list with elements from an existent
        # list and then adds some extra elements (that are going to
        # be journalized) to test the appending of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list.append(4)
        journaled_list.append(5)
        journaled_list.append(6)

        # retrieves the list of appends and tests it
        # against the appended values
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [4, 5, 6])

        # appends one extra value to the list
        # and re-tests the appends list
        journaled_list.append(4)
        self.assertEqual(appends, [4, 5, 6, 4])

        # verifies that the contents of the journaled list
        # are the expected ones
        self.assertEqual(journaled_list, [1, 2, 3, 4, 5, 6, 4])

        # retrieves the removes list, that must be empty and
        # tests it against an empty list
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [])

    def test_remove(self):
        """
        Tests the remove method of the journaled list.
        """

        # creates a journaled list with elements from an existent
        # list and then adds some extra elements (that are going to
        # be journalized) to test the removal of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list.remove(1)
        journaled_list.remove(2)
        journaled_list.remove(3)

        # retrieves the list of removes and tests it
        # against the appended values
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [1, 2, 3])

        # tries to remove the first element one more time
        # (this should raise a value error, and the removes
        # list should remain the same)
        self.assert_raises(ValueError, journaled_list.remove, 1)
        self.assertEqual(removes, [1, 2, 3])

        # verifies that the contents of the journaled list
        # are now empty
        self.assertEqual(journaled_list, [])

        # retrieves the appends list, that must be empty and
        # tests it against an empty list
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [])

    def test_overriding(self):
        """
        Tests the overriding of remove over append operation and
        the reverse.
        """

        # creates a journaled list with elements from an existent
        # list and then adds (and removes) some extra elements (that are going to
        # be journalized) to test the overriding of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list.append(1)
        journaled_list.append(1)
        journaled_list.remove(1)

        # retrieves the list of appends and tests it
        # against the expected values
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [1])

        # retrieves the list of removes and tests it
        # against the expected values
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [])

        # removes one value and tests the appends
        # and removes list (again)
        journaled_list.remove(1)
        self.assertEqual(appends, [])
        self.assertEqual(removes, [])

        # removes one value and tests the appends
        # and removes list (again)
        journaled_list.remove(1)
        self.assertEqual(appends, [])
        self.assertEqual(removes, [1])

        # appends two values and tests the appends
        # and removes list (again)
        journaled_list.append(1)
        journaled_list.append(1)
        self.assertEqual(appends, [1])
        self.assertEqual(removes, [])

    def test_clear_jounal(self):
        """
        Tests the clear journal method of the journaled list.
        """

        # creates a journaled list with elements from an existent
        # list and then adds and removes some extra elements (that are going to
        # be journalized) to test the appending and removal of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list.append(1)
        journaled_list.remove(2)
        journaled_list.remove(3)

        # retrieves the list of appends and tests it
        # against the expected values
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [1])

        # retrieves the list of removes and tests it
        # against the expected values
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [2, 3])

        # verifies that the journaled list contains
        # the expected values
        self.assertEqual(journaled_list, [1, 1])

        # clears the journal and verifies that both
        # the appends and removes lists are now empty
        journaled_list.clear_jounal()
        self.assertEqual(appends, [])
        self.assertEqual(removes, [])

        # verifies that the journaled list remains unmodified
        self.assertEqual(journaled_list, [1, 1])

    def test__append(self):
        """
        Tests the _append method of the journaled list.
        """

        # creates a journaled list with elements from an existent
        # list and then adds some extra elements (that are not going to
        # be journalized) to test the (not jounalized) appending of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list._append(4)
        journaled_list._append(5)
        journaled_list._append(6)

        # retrieves the list of appends and tests it
        # against the empty list (not jounalized)
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [])

        # appends one extra value to the list
        # and re-tests the appends list
        journaled_list._append(4)
        self.assertEqual(appends, [])

        # verifies that the contents of the journaled list
        # are the expected ones
        self.assertEqual(journaled_list, [1, 2, 3, 4, 5, 6, 4])

        # retrieves the removes list, that must be empty and
        # tests it against an empty list
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [])

    def test__remove(self):
        """
        Tests the _remove method of the journaled list.
        """

        # creates a journaled list with elements from an existent
        # list and then adds some extra elements (that are not going to
        # be journalized) to test the (not jounalized) removal of them
        journaled_list = colony.JournaledList([1, 2, 3])
        journaled_list._remove(1)
        journaled_list._remove(2)
        journaled_list._remove(3)

        # retrieves the list of removes and tests it
        # against the empty list (not jounalized)
        removes = journaled_list.get_removes()
        self.assertEqual(removes, [])

        # tries to remove the first element one more time
        # (this should raise a value error, and the removes
        # list should remain the same)
        self.assert_raises(ValueError, journaled_list._remove, 1)
        self.assertEqual(removes, [])

        # verifies that the contents of the journaled list
        # are empty
        self.assertEqual(journaled_list, [])

        # retrieves the appends list, that must be empty and
        # tests it against an empty list
        appends = journaled_list.get_appends()
        self.assertEqual(appends, [])

class OrderedMapTest(colony.ColonyTestCase):
    """
    Class that tests the ordered map structure.
    """

    def test_order(self):
        """
        Tests that the order of item assignment is respected
        in the ordered map structure.
        """

        # creates an ordered map structures with a previously
        # defined sequence of setting of values, to be used
        # to after verify that they are set in the correct
        # order of appearance (test structure)
        ordered_map = colony.OrderedMap()
        ordered_map["1"] = 1
        ordered_map["2"] = 2
        ordered_map["3"] = 3

        # starts the index counter to be used during
        # the iteration part of the test
        index = 0

        # iterates over all the keys in the ordered
        # map to make sure order remains
        for key in ordered_map:
            # verifies that the correct values are found
            # for each of the iteration cycles according
            # to the previously defined map order
            if index == 0: self.assertEqual(key, "1")
            elif index == 1: self.assertEqual(key, "2")
            elif index == 1: self.assertEqual(key, "3")

            # increments the index counter
            # (new iteration)
            index += 1
